
;Shared Machine Code Routines
;============================

;Note that ORG addresses are for illustrative purposes, as actual locations
;vary depending upon the length of each program, each machine code block
;being positioned between the end of BASIC and the start of VARS.


LOCATE: equ  25000           ;CLEAR LOCATE-1
STACKP: equ  23670           ;=SEED =where SP is saved
        org  LOCATE

;adjust any absolute addresses
ADJUST: push bc
        pop  hl              ;HL=actual runtime location of ADJUST
        ld   de,LOCATE       ;DE=assembled      location of ADJUST
        and  a
        sbc  hl,de
        push hl
        pop  bc              ;BC=runtime displacement from LOCATE
        ret  z               ;return if no displacement

PUT_SP: di
        ld   (STACKP),sp     ;save stack pointer
        ld   hl,CHANGE
        add  hl,bc           ;HL=CHANGE+displacement
        ld   sp,hl           ;point SP to list of changes
ADLOOP: pop  hl
        ld   a,h
        or   l
        jr   z,GET_SP

        add  hl,bc           ;HL=location+displacement
        ex   de,hl
        pop  hl
        add  hl,bc           ;HL=content+displacement
        ex   de,hl           ;HL=address of location to change
        ld   (hl),e          ;DE=new content for that location
        inc  hl
        ld   (hl),d
        jr   ADLOOP

GET_SP: ld   sp,(STACKP)     ;restore stack pointer
        ei
        ret

        ;address relocation table
        ;terminated by 0000
        ;part 0: shared routines
        ;ld   hl,$nnnn assembles as 2A nn nn rather than ED 6B nn nn
CHANGE: defw PUT_T+1,SET_W,   GET_T+1,SET_W,    SET_W+1,GET_A, GET_A+4,FIND
        defw GET_A8+1,GET_A,  MCUPDT+1,GET_A8
        defw TO_DO2+1,CHKNEG, TO_DO2+8,CHKNEG,  TO_DO2+15,CHKNEG
        defw MCINIT+1,GET_A8
        defw LABELS0,ADJUST,  LABELS0+2,STACK,  LABELS0+4,PUT_T,  LABELS0+6,GET_T
        defw LABELS0+8,MCMOVE,LABELS0+10,MCUPDT,LABELS0+12,MCINIT,LABELS0+14,COPYAB,LABELS0+16,MCTRAP

        ;part 1: AGES-1 routines
        defw SORT+1,GET_A8,   SORTA+2,JLMT,    SORTB+1,FETROW,   SORTB+5,TMPROW
        defw SORTC+1,FETROW,  SORTC+5,STAKEY,  SORTC+8,TMPROW,   SORTC+11,STAKEY, SORTD+1,TMPROW
        defw MCONDA+1,LOCATS, MCONDA+9,LOCATS ,MCONDA+14,LOCATS, MCONDA+19,LOCATS
        defw MCONDB+1,GETROW, MCONDE+1,MC1250
        defw MC121a+4,MC1250, MCONDF+1,GETROW, MCONDG-2,GETNOL,  MCONDG+1,BASE_X
        defw MC1219+1,GETNOL, MCONDH+1,BASE_X, MC1250+5,MC1213
        defw LOCATS+1,STRING, LOCATS+4,STRING, LOCATS+8,FIND,    GETNOL+8,GETROW
        defw LABELS1,SORT,    LABELS1+2,MCOND1,LABELS1+4,MCOND2, LABELS1+6,SUBSTR

         ;part 2: AGES-2 routines
;        defw INITY0+1,GET_A8,  INDEX30+5,GET_A8, INDEX30+9,GET_A8
;        defw INDEX32+1,FREESP, INDEX32+5,FREESP, LDTEXT+1,TEXT,    LDTEXT+10,OUTVAL, LDTEXT+19,OUTVAL
;        defw SPACEN+1,OUTVAL,  SPACEN+10,OUTVAL, NEXTA+1,ADDLOC,   NEXTJP+1,FORA
;        defw ADD_JP+1,NEXTB,   CALLC2+1,CHUNK2,  CALLI2+1,INDEX2,  NEXTB+7,ADDLOC,   NEXTB+16,ADDLOC
;        defw CHECKY+1,GET_A8,  LDTXNL+1,TEXTNL,  BASRET-2,FORI
;        defw LABELS1,INDEX2,   LABELS1+2, INITY0,LABELS1+4, INDEX30,LABELS1+6, CHECKL

         ;part 3: AGES-3 routines
;        defw NXTARR+2,GET_A8, MCBACK+7,NXTARR, MCBACK+12,NXTINT
;        defw G1224A+1,GO1224, G1224B+1,GO1224, G1222A+1,GO1222, G1222B+1,GO1222
;        defw FTEXT1+1,FCDESC, FTEXT2+1,FFOUND
;        defw NEXTA+1,JPGTLE,  CHEKHS+2,GO1216, NEXTI+1,FORIEM
;        defw MCFRWD+7,NXTARR, MCFRWD+12,NXTINT,GO1244-2,GO1232, NEXTI2+1,GO1232
;        defw MCCOMP+7,NXTARR, MCCOMP+12,NXTINT,NEXTI3+1,FOR1EM
;        defw MCSAVE+7,NXTARR, MCSAVE+12,NXTINT,NEXTI4+1,FOR2K1
;        defw LABELS1,TOPRAM,  LABELS1+2,MCBACK,LABELS1+4,MCFRWD,LABELS1+6,MCSAVE

         ;part 4: AGES-4 routines

        ;end marker
        defw $0000

CLOWER: equ  $0d6e           ;CLS-LOWER clear lower screen
CHOPEN: equ  $1601           ;CHAN-OPEN open channel A
OUTNUM: equ  $1a1b           ;OUT-NUM-1 print integer BC 0-9999
PRINTS: equ  $203c           ;PR-STRING print a string length BC from DE
PLOT:   equ  $22dc           ;PLOT      plot a point
DRAW:   equ  $2477           ;LINE-DRAW draw a line
SCANNG: equ  $24fb           ;the SCANNING subroutine
STORES: equ  $2ab6           ;to calculator stack: AEDCB
FETCHS: equ  $2bf1           ;from calculator stack: AEDCB
STAKA:  equ  $2d28           ;to calculator stack: A
STAKBC: equ  $2d2b           ;to calculator stack: BC
FPTOBC: equ  $2da2           ;from calculator stack: BC
FPTOA:  equ  $2dd5           ;from calculator stack: A

CH_ADD: equ  23645
MCWORK: equ  65368

;stack function parameters passed from BASIC by:
;PRINT USR mcstack,arg1[,arg2...,argn],addr;
STACK:  ld   hl,14
        add  hl,sp
        ld   sp,hl           ;undo USR
        ld   ix,0
GETARG: rst  $20
        call SCANNG
        inc  ix
        rst  $18
        cp   ","
        jr   z,GETARG
        call FPTOBC          ;BC=addr
        ld   h,b
        ld   l,c
        push ix
        pop  bc
        dec  bc
        ld   a,c             ;A =number of parameters
        ld   ix,MCWORK
        jp   (hl)            ;JP addr


;find a string array; call with DE=address of array name
FIND:   ld   bc,(CH_ADD)
        push bc
        ld   (CH_ADD),de
        call $28b2            ;LOOK-VARS
        pop  bc
        ld   (CH_ADD),bc
        ret

;	PRINT USR mcstack,"t$()",t;
;     t$()=array name,
;     t=address of machine code routine
;      =mcputt : put a copy of header array named in t$ into workspace
;      =mcgett : get a copy of header array named in t$ from workspace

;put a copy of header array t$ into workspace
PUT_T:  call SET_W            ;***
        ldir
        ret
;get a copy of header array t$ from workspace
GET_T:  call SET_W            ;***
        ex   de,hl
        ldir
        ret

;set header array & workspace addresses
SET_W:  call GET_A           ;***HL=array address
        ld   de,65536-168    ;DE=workspace address
        ld   bc,54           ;BC=header array size (=48 + 6 byte header)
        ret

;get array address
GET_A:  call FETCHS          ;DE=array name address
        call FIND            ;***HL=array address
        ret
GET_A8: call GET_A          ;***HL=array address
        ld   bc,8
        add  hl,bc          ;HL=address of 1st row in a 2D array
        ret

;get a +ve/-ve integer off the calculator stack
CHKNEG: call FPTOBC
        ret  z               ;return if not -ve
        ld   hl,0000
        and  a
        sbc  hl,bc           ;negate BC
        ld   b,h
        ld   c,l
        ret


;	 PRINT USR mcstack,l,n,o,mcmove;
;     l=length
;     n=new address (higher/lower)
;     o=old address

;move a chunk of the exits data
MCMOVE: call FPTOBC          ;BC=old address
        push bc
        call FPTOBC          ;BC=new address
        push bc
        call FPTOBC          ;BC=length
        pop  de              ;DE=new address
        and  a               ;reset Carry flag
        pop  hl              ;HL=old address
CHUNK2: push hl
        sbc  hl,de           ;old address - new address
        pop  hl
        jr   c,HIGHER
LOWER:  ldir                 ;old address > new address
        ret
HIGHER: add  hl,bc           ;old address < new address
        dec  hl              ;HL=end of old chunk
        push hl
        ex   de,hl
        add  hl,bc
        dec  hl
        ex   de,hl           ;DE=end of new chunk
        pop  hl
        lddr
        ret


;	 PRINT USR mcstack,d,a,s,l,r,f,i$,mcupdt;

;     d=row displacement*rowlen =byte displacement(+/-)
;     a=adjustment(+/-)
;     s=row step*rowlen         =byte step(+/-rowlen)
;     l=stop  row #
;     r=row length
;     f=start row #
;     i$=array name

;for rows f through l of array named in i$
;adjust address pointers by a if greater than p
;shift new row data by (+/-)d bytes from old row location
;step through old row data by (+/-)s bytes

;update a range of index array rows
MCUPDT: call GET_A8          ;***
        push hl          ;a1:>HL=first row address
        call FPTOBC
        push bc          ;a2:>BC=start row #
        call FPTOBC          ;BC=row length
        pop  de          ;a2:<DE=start row #
        pop  hl          ;a1:<HL=first row address
        push bc          ;b1:>BC=row length
        push hl          ;b2:>HL=first row address
        dec  de              ;DE=start row # - 1
        ld   b,c             ;B =row length
        ld   hl,0000
ADD_UP: add  hl,de           ;(start row#-1)*(row length)
        djnz ADD_UP          ;HL=start row offset
        inc  de              ;DE=start row #
        pop  bc          ;b2:<BC=first row address
        add  hl,bc
        push hl          ;c1:>HL=start row address
        push de          ;c2:>DE=start row #
        call FPTOBC          ;BC=stop row #
        pop  hl          ;c2:<HL=start row #
        xor  a
        sbc  hl,bc
        jr   nc,TO_DO        ;jump if start row # >= stop row #
        ld   d,a             ;sign -ve
        ld   e,a             ;DE=0000
        ex   de,hl
        and  a
        sbc  hl,de           ;reverse sign to +ve
TO_DO:  inc  hl
        pop  de          ;c1:<DE=start row address
        push hl          ;d1:>HL=number of rows to do
        exx
        pop  bc          ;d1:<BC'=number of rows to do
        exx
        push de          ;b2:>DE=start row address
TO_DO2: call CHKNEG          ;***BC=row step*rowlen =byte step(+/-rowlen)
        ld   (MCWORK+75),bc
        call CHKNEG          ;***BC=adjustment(+/-)
        ld   (MCWORK+77),bc
        call CHKNEG          ;***BC=row displacement*rowlen(+/-)
        ld   (MCWORK+79),bc
        pop  hl          ;b2:<HL=start row address
NEXT:   ld   e,(hl)          ;HL=old row address
        inc  hl
        ld   d,(hl)          ;DE=old pointer
        ld   a,d
        or   e
        jr   z,NOADD         ;don't update if old pointer is zero
TO_DO6: ld   bc,(MCWORK+77)  ;BC=adjustment
        ex   de,hl           ;HL=old pointer
        add  hl,bc           ;HL=old pointer + adjustment
        ex   de,hl           ;DE=new pointer
NOADD:  ld   (hl),d
        dec  hl
        ld   (hl),e          ;put new pointer at old address
        ex   de,hl           ;DE=old row address
TO_DO7: ld   hl,(MCWORK+79)  ;HL=displacement
        add  hl,de           ;HL=old row address + displacement (may be 0)
        ex   de,hl           ;HL=old row address; DE=new row address
        pop  bc          ;b1:<BC=row length
        push bc          ;b1:>BC=row length
        push hl          ;b2:>HL=old row address
        ldir                 ;copy updated row to new address
        pop  hl          ;b2:<HL=old row address
TO_DO5: ld   bc,(MCWORK+75)  ;BC=step
        add  hl,bc           ;HL=old address + step
        exx
        dec  bc              ;decrement row counter
        ld   a,b
        or   c
        exx
        jr   nz,NEXT
        pop  bc          ;b1:<BC=row length
        ret


;PRINT USR mcstack,l,x,n$,mcinitn;
MCINIT: call GET_A8         ;***
        push hl             ;HL=address of 1st row in array
        call FPTOBC         ;BC=number of rows
        push bc
        call FPTOA          ;A =row length
        ld   b,a
        ld   c,a
        pop  de             ;DE=number of rows
        pop  hl
        push hl             ;save original array pointer
REPS:   ld   (hl),$00       ;initialise 1st row
        inc  hl
        djnz REPS
        push hl             ;save updated array pointer
        dec  de             ;DE=(number of rows)-1 for row already done
        ld   hl,0000
        ld   b,c
REPD:   add  hl,de          ;multiply (row#-1)*rowlen
        djnz REPD
        ld   b,h
        ld   c,l            ;BC=number of bytes to copy
        pop  de             ;DE=starting destination (1+rowlen byte of array)
        pop  hl             ;HL=starting source      (1st byte of array)
        ldir                ;initialise remainder of array rows
        ret


;PRINT USR mcstack,a,b,l[,"c"],copy;
;[optionally copy string 'c' to 'a' then] copy 'l' bytes from 'a' to 'b' (may overlap)
COPYAB: cp   4
        jr   z,COPY2
        xor  a
        call STAKA      ;stack zero for an empty string
COPY2:  call FETCHS     ;get initialisation string
        push bc      ;1>;BC=string length
        push de      ;2>;DE=string address
        call FPTOBC
        pop  de      ;2<
        pop  hl      ;1<
        push bc      ;1>;BC=copy length
        push hl      ;2>
        push de      ;3>
        call FPTOBC
        pop  de      ;3<
        pop  hl      ;2<
        push bc      ;2>;BC=copy to address
        push hl      ;3>
        push de      ;4>
        call FPTOBC
        ld   d,b
        ld   e,c        ;DE=copy from address
        pop  hl      ;4<;HL=initialisation string address
        pop  bc      ;3<;BC=initialisation string length
        push de      ;3>
        ld   a,b
        or   c
        jr   z,COPY3
        ldir            ;copy initialisation string
        scf
COPY3:  pop  hl      ;3<
        pop  de      ;2<
        pop  bc      ;1<
        jr   c,COPYLO    ;jump if init string set
        push hl
        sbc  hl,de      ;from - to
        pop  hl
        jr   c,COPYHI
COPYLO: ldir            ;from > to -or- initialisation string specified
        ret
COPYHI: add  hl,bc      ;from < to
        dec  hl         ;HL=end of from chunk
        push hl
        ex   de,hl
        add  hl,bc
        dec  hl
        ex   de,hl      ;DE=end of to chunk
        pop  hl
        lddr
        ret


;ERROR TRAP - machine code segment

MCTRAP: RES  4,(IY+1)       ;patch for +3 128k mode

        LD   A,(23621)      ;save PPC line number (where error occurred)
        LD   (23615),A      ;in LIST_SP
        LD   A,(23622)
        LD   (23616),A
        LD   A,(23610)      ;save ERR_NR error code
        LD   (23560),A      ;in LAST_K
        LD   BC,9000        ;new line number to jump to
        LD   (23662),BC     ;OLDPPC - CONTINUE line number
        LD   A,$ff
        LD   (23610),A      ;ERR_NR - reset error code
        INC  A
        LD   (23664),A      ;OSPCC  - CONTINUE statement number
        LD   HL,(23641)     ;E_LINE - address of editor buffer
        CALL $1652          ;ONE-SPACE - make a space in it
        EX   DE,HL
        LD   (HL),$e8       ;insert code for "CONTINUE"
MCLOOP: CALL $1f54          ;BREAK-KEY
        JP   C,$12cf        ;MAIN-3 - jump if not BREAK
        JR   MCLOOP         ;ignore BREAK

;++++++++++++++++++++++++++++++