
;Shared Machine Code Routines
;============================

;Note that ORG addresses are for illustrative purposes, as actual locations
;vary depending upon the length of each program, each machine code block
;being positioned between the end of BASIC and the start of VARS.


LOCATE: equ  25000           ;CLEAR LOCATE-1
STACKP: equ  23670           ;=SEED =where SP is saved
        org  LOCATE

;adjust any absolute addresses
ADJUST: push bc
        pop  hl              ;HL=actual runtime location of ADJUST
        ld   de,LOCATE       ;DE=assembled      location of ADJUST
        and  a
        sbc  hl,de
        push hl
        pop  bc              ;BC=runtime displacement from LOCATE
        ret  z               ;return if no displacement

PUT_SP: di
        ld   (STACKP),sp     ;save stack pointer
        ld   hl,CHANGE
        add  hl,bc           ;HL=CHANGE+displacement
        ld   sp,hl           ;point SP to list of changes
ADLOOP: pop  hl
        ld   a,h
        or   l
        jr   z,GET_SP

        add  hl,bc           ;HL=location+displacement
        ex   de,hl
        pop  hl
        add  hl,bc           ;HL=content+displacement
        ex   de,hl           ;HL=address of location to change
        ld   (hl),e          ;DE=new content for that location
        inc  hl
        ld   (hl),d
        jr   ADLOOP

GET_SP: ld   sp,(STACKP)     ;restore stack pointer
        ei
        ret

        ;address relocation table
        ;terminated by 0000
        ;part 0: shared routines
        ;ld   hl,$nnnn assembles as 2A nn nn rather than ED 6B nn nn
CHANGE: defw FINDAR+1,GET_A,  PUT_T+1,SET_W,   GET_T+1,SET_W,    SET_W+1,GET_A, GET_A+4,FIND
        defw GET_A6+1,GET_A,  GET_A8+1,GET_A,  MCUPDT+1,GET_A8
        defw TO_DO2+1,CHKNEG, TO_DO2+8,CHKNEG,  TO_DO2+15,CHKNEG
        defw CALL2+1,FINDAR
        defw LABELS0,ADJUST,  LABELS0+2,STACK,  LABELS0+4,PUT_T,  LABELS0+6,GET_T
        defw LABELS0+8,MCMOVE,LABELS0+10,MCUPDT,LABELS0+12,FILL,LABELS0+14,COPYAB,LABELS0+16,MCTRAP

        ;part 1: AGES-1 routines
        defw SORT+1,GET_A8,   SORTB+1,FETROW
        defw SORTC+1,FETROW,  SORTC+5,STAKEY,  SORTC+11,STAKEY
        defw MCONDA+1,LOCATS, MCONDA+9,LOCATS ,MCONDA+14,LOCATS, MCONDA+19,LOCATS
        defw MCONDB+1,GETROW, MCONDE+1,MC1250
        defw MC121a+4,MC1250, MCONDF+1,GETROW, MCONDG-2,GETNOL,  MCONDG+1,BASE_X
        defw MC1219+1,GETNOL, MCONDH+1,BASE_X, MC1255+1,MC1213
        defw LOCATS+1,STRING, LOCATS+4,STRING, LOCATS+8,FIND,    GETNOL+8,GETROW
        defw SWAPRG+1,GET_A,  SWAPRG+5,GET_A,  EXPG2R+1,CLEARR
        defw LABELS1,SORT,    LABELS1+2,MCOND1,LABELS1+4,MCOND2, LABELS1+6,SUBSTR
        defw LABELS1+8,SWAPRG

        ;end marker
        defw $0000

CLOWER: equ  $0d6e           ;CLS-LOWER clear lower screen
CHOPEN: equ  $1601           ;CHAN-OPEN open channel A
OUTNUM: equ  $1a1b           ;OUT-NUM-1 print integer BC 0-9999
PRINTS: equ  $203c           ;PR-STRING print a string length BC from DE
PLOT:   equ  $22dc           ;PLOT      plot a point
DRAW:   equ  $2477           ;LINE-DRAW draw a line
SCANNG: equ  $24fb           ;the SCANNING subroutine
STORES: equ  $2ab6           ;to calculator stack: AEDCB
FETCHS: equ  $2bf1           ;from calculator stack: AEDCB
STAKA:  equ  $2d28           ;to calculator stack: A
STAKBC: equ  $2d2b           ;to calculator stack: BC
FPTOBC: equ  $2da2           ;from calculator stack: BC
FPTOA:  equ  $2dd5           ;from calculator stack: A

CH_ADD: equ  23645
MCWORK: equ  65368

;stack function parameters passed from BASIC by:
;PRINT USR mcstack,arg1[,arg2...,argn],addr;
STACK:  ld   hl,14
        add  hl,sp
        ld   sp,hl           ;undo USR
        ld   ix,0
GETARG: rst  $20
        call SCANNG
        inc  ix
        rst  $18
        cp   ","
        jr   z,GETARG
        call FPTOBC          ;BC=addr
        ld   h,b
        ld   l,c
        push ix
        pop  bc
        dec  bc
        ld   a,c             ;A =number of parameters
        ld   ix,MCWORK
        jp   (hl)            ;JP addr


;find a string array; call with DE=address of array name
FIND:   ld   bc,(CH_ADD)
        push bc
        ld   (CH_ADD),de
        call $28b2      ;LOOK-VARS
                        ;if found: Carry reset, HL=address; otherwise: Carry set
        ld   a,c        ;A =discriminator byte; bits 6&5 indicate variable type
        pop  bc
        ld   (CH_ADD),bc
        ret
FINDAR: call GET_A      ;***find array address
        ret  c          ;return if not found
        push hl         ;HL=address of array
        inc  hl
        ld   e,(hl)
        inc  hl
        ld   d,(hl)     ;DE=1+ length of dims+elements
        inc  hl
        ld   b,(hl)     ;B =number of dimensions
        inc  hl
        dec  de
ADDUPD: inc  hl
        dec  de
        inc  hl
        dec  de
        djnz ADDUPD
        ld   b,d
        ld   c,e        ;BC=total size of array elements
        ex   de,hl      ;DE=address of array's 1st element
        pop  hl         ;HL=address of array
        and  a          ;reset Carry
        ret

;	PRINT USR mcstack,"t$()",t;
;     t$()=array name,
;     t=address of machine code routine
;      =mcputt : put a copy of header array named in t$ into workspace
;      =mcgett : get a copy of header array named in t$ from workspace

;put a copy of header array t$ into workspace
PUT_T:  call SET_W            ;***
        ldir
        ret
;get a copy of header array t$ from workspace
GET_T:  call SET_W            ;***
        ex   de,hl
        ldir
        ret

;set header array & workspace addresses
SET_W:  call GET_A           ;***HL=array address
        ld   de,65536-168    ;DE=workspace address
        ld   bc,54           ;BC=header array size (=48 + 6 byte header)
        ret

;get array address
GET_A:  call FETCHS          ;DE=array name address
        call FIND            ;***HL=array address
        ret
GET_A6: call GET_A          ;***HL=array address
        ld   bc,6
        add  hl,bc          ;HL=address of 1st row in a 1D array
        ret
GET_A8: call GET_A          ;***HL=array address
        ld   bc,8
        add  hl,bc          ;HL=address of 1st row in a 2D array
        ret

;get a +ve/-ve integer off the calculator stack
CHKNEG: call FPTOBC
        ret  z               ;return if not -ve
        ld   hl,0000
        and  a
        sbc  hl,bc           ;negate BC
        ld   b,h
        ld   c,l
        ret


;	 PRINT USR mcstack,l,n,o,mcmove;
;     l=length
;     n=new address (higher/lower)
;     o=old address

;move a chunk of the exits data
MCMOVE: call FPTOBC          ;BC=old address
        push bc
        call FPTOBC          ;BC=new address
        push bc
        call FPTOBC          ;BC=length
        pop  de              ;DE=new address
        and  a               ;reset Carry flag
        pop  hl              ;HL=old address
CHUNK2: push hl
        sbc  hl,de           ;old address - new address
        pop  hl
        jr   c,HIGHER
LOWER:  ldir                 ;old address > new address
        ret
HIGHER: add  hl,bc           ;old address < new address
        dec  hl              ;HL=end of old chunk
        push hl
        ex   de,hl
        add  hl,bc
        dec  hl
        ex   de,hl           ;DE=end of new chunk
        pop  hl
        lddr
        ret


;	 PRINT USR mcstack,d,a,s,l,r,f,i$,mcupdt;

;     d=row displacement*rowlen =byte displacement(+/-)
;     a=adjustment(+/-)
;     s=row step*rowlen         =byte step(+/-rowlen)
;     l=stop  row #
;     r=row length
;     f=start row #
;     i$=array name

;for rows f through l of array named in i$
;adjust address pointers by a if greater than p
;shift new row data by (+/-)d bytes from old row location
;step through old row data by (+/-)s bytes

;update a range of index array rows
MCUPDT: call GET_A8          ;***
        push hl          ;a1:>HL=first row address
        call FPTOBC
        push bc          ;a2:>BC=start row #
        call FPTOBC          ;BC=row length
        pop  de          ;a2:<DE=start row #
        pop  hl          ;a1:<HL=first row address
        push bc          ;b1:>BC=row length
        push hl          ;b2:>HL=first row address
        dec  de              ;DE=start row # - 1
        ld   b,c             ;B =row length
        ld   hl,0000
ADD_UP: add  hl,de           ;(start row#-1)*(row length)
        djnz ADD_UP          ;HL=start row offset
        inc  de              ;DE=start row #
        pop  bc          ;b2:<BC=first row address
        add  hl,bc
        push hl          ;c1:>HL=start row address
        push de          ;c2:>DE=start row #
        call FPTOBC          ;BC=stop row #
        pop  hl          ;c2:<HL=start row #
        xor  a
        sbc  hl,bc
        jr   nc,TO_DO        ;jump if start row # >= stop row #
        ld   d,a             ;sign -ve
        ld   e,a             ;DE=0000
        ex   de,hl
        and  a
        sbc  hl,de           ;reverse sign to +ve
TO_DO:  inc  hl
        pop  de          ;c1:<DE=start row address
        push hl          ;d1:>HL=number of rows to do
        exx
        pop  bc          ;d1:<BC'=number of rows to do
        exx
        push de          ;b2:>DE=start row address
TO_DO2: call CHKNEG          ;***BC=row step*rowlen =byte step(+/-rowlen)
        ld   (MCWORK+75),bc
        call CHKNEG          ;***BC=adjustment(+/-)
        ld   (MCWORK+77),bc
        call CHKNEG          ;***BC=row displacement*rowlen(+/-)
        ld   (MCWORK+79),bc
        pop  hl          ;b2:<HL=start row address
NEXT:   ld   e,(hl)          ;HL=old row address
        inc  hl
        ld   d,(hl)          ;DE=old pointer
        ld   a,d
        or   e
        jr   z,NOADD         ;don't update if old pointer is zero
TO_DO6: ld   bc,(MCWORK+77)  ;BC=adjustment
        ex   de,hl           ;HL=old pointer
        add  hl,bc           ;HL=old pointer + adjustment
        ex   de,hl           ;DE=new pointer
NOADD:  ld   (hl),d
        dec  hl
        ld   (hl),e          ;put new pointer at old address
        ex   de,hl           ;DE=old row address
TO_DO7: ld   hl,(MCWORK+79)  ;HL=displacement
        add  hl,de           ;HL=old row address + displacement (may be 0)
        ex   de,hl           ;HL=old row address; DE=new row address
        pop  bc          ;b1:<BC=row length
        push bc          ;b1:>BC=row length
        push hl          ;b2:>HL=old row address
        ldir                 ;copy updated row to new address
        pop  hl          ;b2:<HL=old row address
TO_DO5: ld   bc,(MCWORK+75)  ;BC=step
        add  hl,bc           ;HL=old address + step
        exx
        dec  bc              ;decrement row counter
        ld   a,b
        or   c
        exx
        jr   nz,NEXT
        pop  bc          ;b1:<BC=row length
        ret


;PRINT USR stack,"a","b",fill;
;fill array 'a' elements with value 'b'
MEMBOT: equ  23698

FILL:   rst  $28
        defb $01,$38    ;exchange top two calculator stack items
CALL2:  call FINDAR     ;***
        jp   c,$1C2E    ;"2 Variable not found"
        push bc      ;1>;BC=total size of array elements
        push de      ;2>;DE=address of array's 1st element
        push af      ;3>;A =discriminator byte; bits 6&5 indicate variable type
        call FETCHS     ;get fill value parameters
        pop  hl      ;3<
        bit  6,h        ;H =discriminator byte
        jr   nz,FILL2   ;jump if character array
        ld   hl,MEMBOT+4
        ld   (hl),b
        dec  hl
        ld   (hl),c
        dec  hl
        ld   (hl),d
        dec  hl
        ld   (hl),e
        dec  hl
        ld   (hl),a
        ld   bc,5
        ex   de,hl
FILL2:  ex   de,hl      ;HL=fill value address
        pop  de      ;2<;DE=address of array's 1st element
        push de      ;2>
        push bc      ;3>
        ldir            ;copy fill value to 1st element
        pop  bc      ;3<;BC=size of fill value
        pop  af      ;2<;AF=address of array's 1st element
        pop  hl      ;1<;HL=total size of array elements
        push af      ;1>
        and  a
        sbc  hl,bc
        ld   b,h
        ld   c,l        ;BC=size of unfilled segment
        pop  hl      ;1<;HL=address of array's 1st element
        ldir
        ret


;PRINT USR mcstack,a,b,l[,"c"],copy;
;[optionally copy string 'c' to 'a' then] copy 'l' bytes from 'a' to 'b' (may overlap)
COPYAB: cp   4
        jr   z,COPY2
        xor  a
        call STAKA      ;stack zero for an empty string
COPY2:  call FETCHS     ;get initialisation string
        push bc      ;1>;BC=string length
        push de      ;2>;DE=string address
        call FPTOBC
        pop  de      ;2<
        pop  hl      ;1<
        push bc      ;1>;BC=copy length
        push hl      ;2>
        push de      ;3>
        call FPTOBC
        pop  de      ;3<
        pop  hl      ;2<
        push bc      ;2>;BC=copy to address
        push hl      ;3>
        push de      ;4>
        call FPTOBC
        ld   d,b
        ld   e,c        ;DE=copy from address
        pop  hl      ;4<;HL=initialisation string address
        pop  bc      ;3<;BC=initialisation string length
        push de      ;3>
        ld   a,b
        or   c
        jr   z,COPY3
        ldir            ;copy initialisation string
        scf
COPY3:  pop  hl      ;3<
        pop  de      ;2<
        pop  bc      ;1<
        jr   c,COPYLO    ;jump if init string set
        push hl
        sbc  hl,de      ;from - to
        pop  hl
        jr   c,COPYHI
COPYLO: ldir            ;from > to -or- initialisation string specified
        ret
COPYHI: add  hl,bc      ;from < to
        dec  hl         ;HL=end of from chunk
        push hl
        ex   de,hl
        add  hl,bc
        dec  hl
        ex   de,hl      ;DE=end of to chunk
        pop  hl
        lddr
        ret


;ERROR TRAP - machine code segment

MCTRAP: RES  4,(IY+1)       ;patch for +3 128k mode

        LD   A,(23621)      ;save PPC line number (where error occurred)
        LD   (23615),A      ;in LIST_SP
        LD   A,(23622)
        LD   (23616),A
        LD   A,(23610)      ;save ERR_NR error code
        LD   (23560),A      ;in LAST_K
        LD   BC,9000        ;new line number to jump to
        LD   (23662),BC     ;OLDPPC - CONTINUE line number
        LD   A,$ff
        LD   (23610),A      ;ERR_NR - reset error code
        INC  A
        LD   (23664),A      ;OSPCC  - CONTINUE statement number
        LD   HL,(23641)     ;E_LINE - address of editor buffer
        CALL $1652          ;ONE-SPACE - make a space in it
        EX   DE,HL
        LD   (HL),$e8       ;insert code for "CONTINUE"
MCLOOP: CALL $1f54          ;BREAK-KEY
        JP   C,$12cf        ;MAIN-3 - jump if not BREAK
        JR   MCLOOP         ;ignore BREAK

;++++++++++++++++++++++++++++++

;   1 REM Shell Sort
;  10 LET n=80: DIM v(n): FOR i=1 TO n: LET v(i)=INT (RND*100): NEXT i
;  20 FOR i=1 TO n: LET a=2^i: IF a<n THEN NEXT i
;  30 LET m=a/2-1
;  35 LET ir=n-m
;  40 FOR s=1 TO ir: LET j=s
;  45 LET k=j+m: IF v(j)<v(k) THEN GO TO 300
;  50 LET l=v(j): LET v(j)=v(k): LET v(k)=l: LET j=j-m: IF j>0 THEN GO TO 45
; 300 NEXT s
; 310 IF m=1 THEN FOR i=1 TO n: PRINT v(i): NEXT i: STOP
; 320 LET m=(m-1)/2: GO TO 35

;    PRINT USR mcstack,offset,"c$()",mcsort;
;     --------------------
CROW:   equ  300
CROLEN: equ  28              ;c$() row length
MSTART: equ  255
J_:     equ  MCWORK   ;defs 2;(IX+0)=j;
L_:     equ  MCWORK+2 ;defs 2;(IX+2)=e;
M_:     equ  MCWORK+4 ;defs 2;(IX+4)=m;
K_:     equ  MCWORK+6 ;defs 2;(IX+6)=k;
LH:     equ  MCWORK+8 ;defs 2;(IX+8)=c$() base address;
TMPROW: equ  LH+2     ;defs CROLEN;(IX+10)=c$(j)

SORT:   call GET_A8          ;***
        ld   (LH),hl         ;HL=address of 1st row in c$()
        call FPTOBC          ;BC=key offset in row: 0,2,4,6
SORTA:  ld   (L_),bc
        ld   de,MSTART
        ld   (M_),de         ;  =m start value

LINE35: ld   hl,CROW         ;HL=n number of rows
        ld   bc,(M_)
        and  a
        sbc  hl,bc
        ex   de,hl           ;DE=ir=crow-m
        ld   bc,0
LINE40: inc  bc              ;BC=s
        push bc
        ld   (J_),bc         ;  =j=s             ROW j INDEX
LINE45: ld   hl,(J_)
        ld   bc,(M_)
        add  hl,bc           ;HL=t=j+m
        ld   (K_),hl         ;                   ROW k INDEX
        push de
        ld   bc,(J_)         ;BC=j               ROW j INDEX
SORTB:  call FETROW          ;***
        push hl              ;HL=c$(j) address
        ld   de,TMPROW
        ld   bc,CROLEN
        ldir                 ;  =l$=c$(j)        TMPROW=c$(j)
        ld   bc,(K_)         ;BC=t               ROW k INDEX
SORTC:  call FETROW          ;***
        push hl              ;HL=c$(k) address
        call STAKEY          ;***
        ld   hl,TMPROW
        call STAKEY          ;***
        ld   b,$12
        rst  $28             ;start calculator
        defb $12             ;x$>=y$
        defb $38
        call FPTOA           ;A =TRUE or FALSE
        and  a
        jr   nz,NEXTS1       ;jump if c$(k)>=c$(j)
SWITCH: pop  hl              ;get c$(k) address
        pop  de              ;get c$(j) address
        push hl              ;save c$(k) address
        ld   bc,CROLEN
        ldir
SORTD:  ld   hl,TMPROW
        pop  de              ;get c$(k) address
        ld   bc,CROLEN
        ldir
        ld   hl,(M_)         ;  =m
        ld   bc,(J_)
        and  a
        sbc  hl,bc           ;  =m-j
        jr   nc,NEXTS2       ;if m>=j
        ex   de,hl
        ld   hl,0
        and  a
        sbc  hl,de           ;  =j-m
        ld   (J_),hl         ; j=j-m
        pop  de
        jr   LINE45
JRBACK: jr   LINE35
NEXTS1: pop  hl
        pop  hl
NEXTS2: pop  de
        pop  bc
        dec  de
        ld   a,d
        or   e
        jr   nz,LINE40
        ld   hl,1
        ld   bc,(M_)
        sbc  hl,bc
        ret  z               ;ret if m=1
        dec  bc
        srl  b
        rr   c
        ld   (M_),bc         ;m=(m-1)/2
        jr   JRBACK

FETROW: dec  bc              ;BC=row index-1
        call STAKBC
        ld   bc,CROLEN
        call STAKBC
        rst  $28             ;start calculator
        defb $04             ;multiply           (ROW INDEX-1)*(ROW LENGTH)
        defb $38
        call FPTOBC          ;BC=row offset
        ld   hl,(LH)         ;HL=                BASE ADDRESS
        add  hl,bc           ;                   ROW ADDRESS
        ret

STAKEY: ld   de,(L_)
        add  hl,de
        ex   de,hl           ;DE=c$(?) key address
        ld   bc,19           ;BC=c$(?) key length
        call STORES
        ret


;++++++++++++++++++++++++++++++


;PRINT USR ms,k$,b1,crow,mn1;

;set up n$() conditions array & data entries
;initial entry point
MCOND1: ld   hl,MCWORK
        push hl
        pop  ix
        xor  a
        ld   (hl),a
        push hl
        pop  de
        inc  de
        ld   bc,40
        ldir                 ;reset workspace

        call FPTOBC
        ld   (MCWORK+37),bc  ;=cdec =decrementing counter of c$() rows processed
        ld   (MCWORK+33),bc  ;=crow =number of rows in c$()
        call FPTOBC
        ld   (MCWORK+19),bc  ;=b1 =conditions data base address
        call FETCHS
        ld   (ix+29),c       ;BC=k$ length
        ex   de,hl           ;HL=k$ address
        ld   de,MCWORK+39
        ldir                 ;copy quantity keys list

        ld   bc,6            ;BC=length of 1.dim array header
        ld   a,"b"
MCONDA: CALL LOCATS          ;***find b$()
        ld   bc,8            ;BC=length of 2.dim array header
        ld   a,"c"
        CALL LOCATS          ;***find c$()
        ld   a,"i"
        call LOCATS          ;***find i$()

        ld   a,"n"
        call LOCATS          ;***find n$()
        ld   ix,MCWORK       ;reset IX

        ld   bc,500          ;BC=rows in i$()
        ld   de,6            ;DE=row length
        ld   hl,(MCWORK+4)   ;HL=i$() base address
        inc  hl
        inc  hl
        inc  hl              ;HL=address of i$(1,4) n$() key
NEXTNK: ld   (hl),0          ;reset i$(?,4) n$ key
        add  hl,de
        dec  bc
        ld   a,b
        or   c
        jr   nz,NEXTNK

MC1213: inc  bc               ;BC=i
        PUSH bc
        ld   hl,(MCWORK+2)   ;HL=c$() base
        ld   a,CROLEN        ;A =row length
MCONDB: call GETROW          ;***get c$(i) address
        ld   (MCWORK+8),hl   ;=c$(i) address
        ld   bc,6
        add  hl,bc           ;skip over location numbers

        xor  a
        ld   c,(ix+29)       ;C =quantity keys list length
        ld   b,a
        ld   (ix+30),a       ;  =quantity keys list key (reset)
        ld   e,(hl)
        inc  hl
        ld   d,(hl)          ;DE=1st 2 chars of description
        ld   (MCWORK+31),de
        ld   hl,MCWORK+39    ;HL=quantity keys list address
        add  hl,bc
        ld   b,c             ;B =quantity keys list length

LISTAD: dec  hl
        ld   d,(hl)
        dec  hl
        ld   e,(hl)
        push de              ;stack list of quantity key values
        dec  b
        djnz LISTAD

        ld   b,c             ;B =quantity keys list length
        ld   a,b
        ld   de,(MCWORK+31)  ;DE=1st 2 chars of description
MATCH3: pop  hl
        and  a
        sbc  hl,de           ;check description prefix
        jr   nz,MATCHZ
        sub  b               ;set key if a match
        sra  a
        ld   (ix+30),a       ;  =quantity keys list key
MATCHZ: dec  b
        djnz MATCH3          ;repeat until all key values unstacked
        or a
MCONDE: jp   z,MC1250        ;***skip if description is blank

;check for an incomplete entry
MC1214: ld   b,3             ;A =loop counter
        ld   hl,(MCWORK+8)   ;HL=c$(i) address
MC121a: ld   a,(hl)
        inc  hl
        or   (hl)
        jp   z,MC1250        ;***entry=0
        inc  hl
        djnz MC121a

        ;check c$() sorted on col.1 (From)
MC1216: ld   hl,(MCWORK+8)   ;HL=c$(i) address
        ld   b,(hl)
        inc  hl
        ld   c,(hl)          ;BC=c$(i,TO k2) hi-lo (for sorting)
        ld   (MCWORK+14),bc  ;loc=c$(i,TO k2)
        PUSH bc
        POP  hl              ;HL=loc
        ld   de,(MCWORK+16)  ;DE=poc
        and  a
        sbc  hl,de
        jr   nc,MC1217

        pop   af
        ld   bc,1620         ;"CON.SORT<>Col.1" fatal error
        ld   (MCWORK+27),bc
        ret                  ;return to BASIC

        ;check for any conditions at this location
MC1217: ld   (MCWORK+16),bc  ;poc=loc
        ld   hl,(MCWORK+4)   ;HL=i$() base address
        ld   a,6             ;A =i$() row length
        ld   bc,(MCWORK+14)  ;BC=loc
MCONDF: CALL GETROW          ;***
        inc  hl
        inc  hl
        inc  hl
        xor  a
        cp   (hl)
        jr   z,NONROW        ;jump if no i$() entry

        ld   c,(hl)          ;C =n$() (location/data x-ref) key
MC1219: call GETNOL          ;***HL=n$(nkey) address
        jr   MC1220

NONROW: inc  (ix+18)         ;nkey=nkey+k1
        ld   c,(ix+18)       ;C =n$() (location/data x-ref) key
        ld   (hl),c          ;i$(loc,4)=n$() (location/data x-ref) key
        call GETNOL          ;***HL=n$(nkey) address
MCONDG: call BASE_X          ;***DE=next entry address
        ld   (hl),e
        inc  hl
        ld   (hl),d          ;n$(nkey)=FN i$(b1+clen,k0)
        dec  hl

MC1220: ld   (MCWORK+12),hl
        ld   e,(hl)          ;get data start & length from n$(nkey)
        inc  hl
        ld   d,(hl)
        inc  hl
        ld   c,(hl)
        ex   de,hl           ;HL=start
        ld   b,0             ;BC=length
        add  hl,bc           ;HL=address to put next data entry
        ld   (MCWORK+24),hl  ;jpos=sta+len=address to put next data entry
MCONDH: call BASE_X          ;***DE=b1+clen=current top of used data area
        and  a
        sbc  hl,de
        jr   nc,MC1222       ;jump if address is free

        POP  bc              ;otherwise return to make room & update n$()
        ld   (MCWORK+35),bc  ;=i
        ld   bc,1640         ;GOTO line number
        ld   (MCWORK+27),bc
        ret                  ;return to BASIC for:
          ;PRINT USR ms,b1+clen-jpos,jpos+k6,jpos,mm;:
          ;PRINT USR ms,k0,k6,k3,nrow,k3,CODE i$(loc,k4)+k1,"n$()",mu;

;PRINT USR ms,mn2;

;re-entry point after MC1220 exit
MCOND2: ld   ix,MCWORK
        ld   bc,(MCWORK+35)  ;BC=i
        PUSH bc
MC1222: ld   de,(MCWORK+8)   ;DE=c$(i) address
        ld   b,2             ;B = loop counter
        ld   hl,(MCWORK+24)  ;HL=jpos

MC1223: PUSH  bc
        inc  de
        inc  de              ;DE=c$(i,e+2) address
        PUSH de
        PUSH hl
        ld   a,(de)
        ld   b,a
        inc  de
        ld   a,(de)
        ld   c,a             ;BC=c$(i,e TO e+1) hi-lo
        POP  hl              ;HL=jpos
        ld   (hl),c          ;POKE jpos,FN g(s)
        inc  hl
        ld   (hl),b          ;POKE jpos+k1,FN h(s)
        inc  hl
        POP  de              ;DE=c$(i,e) address
        POP  bc              ;B =loop counter
        djnz MC1223

MC1225: POP  bc              ;BC=i
        ld   (hl),c
        inc  hl
        ld   (hl),b          ;  =c$() key
        PUSH bc
        ld   hl,(MCWORK+21)  ;HL=clen
        ld   bc,6
        add  hl,bc
        ld   (MCWORK+21),hl  ;clen=clen+k6
        ld   de,(MCWORK+12)  ;DE=n$(nkey) address
        inc  de
        inc  de
        ld   a,(de)          ;A =n$(nkey,k3)
        add  a,6
        ld   (de),a          ;n$(nkey,k3)=CHR$ (len+k6)

;jump back if any more c$() rows to do
MC1250: POP  bc              ;BC=i
        ld   de,(MCWORK+37)
        dec  de              ;decrement crow loop counter
        ld   (MCWORK+37),de
        ld   a,d
        or   e
MC1255: jp   nz,MC1213       ;***

        ld   bc,1660         ;GOTO line number (finished)
        ld   (MCWORK+27),bc
        ret                  ;return to BASIC

STRING: defm "?$()"

LOCATS: ld   (STRING),a      ;***A =array name
        ld   de,STRING       ;***DE=location of array name
        push bc
        call FIND            ;***HL=array header address
        pop  bc              ;BC=array header length
        add  hl,bc           ;HL=array data address
        ld   (ix+0),l
        ld   (ix+1),h
        inc  ix
        inc  ix
        ret

GETROW: push hl
        push af
        dec  bc              ;BC=index-1
        call STAKBC
        pop  af              ;A =length
        call STAKA
        rst  $28             ;start calc
        defb $04             ;(index-1)*length
        defb $38             ;end calc
        call FPTOBC          ;BC=array(index) row offset
        pop  hl              ;HL=array base address
        add  hl,bc           ;HL=array(index) row address
        ret

;get n$(key) address
GETNOL: ld   hl,(MCWORK+6)   ;HL=n$() base address
        ld   b,0             ;BC=key set prior to call
        ld   a,3             ;A =row length
        CALL GETROW          ;***get n$(nkey) address
        ret

;calculate b1+clen
BASE_X: push hl
        ld   hl,(MCWORK+19)  ;HL=b1=conditions data base address
        ld   de,(MCWORK+21)  ;DE=clen=conditions data length
        add  hl,de
        ex   de,hl           ;DE=next entry address
        pop  hl
        ret


;       org  MCWORK
;       defw 0000            ;(ix+00) address of b$() row
;       defw 0000            ;(ix+02) address of first c$() row
;       defw 0000            ;(ix+04) address of first i$() row
;       defw 0000            ;(ix+06) address of first n$() row
;       defw 0000            ;(ix+08) address of current c$() row
;       defw 0000            ;(ix+10) address of current i$() row
;       defw 0000            ;(ix+12) address of current n$() row
;       defw 0000            ;(ix+14) loc     current  c$() From location
;       defw 0000            ;(ix+16) poc     previous c$() From location
;       defb 00              ;(ix+18) nkey    last used n$() key
;       defw 0000            ;(ix+19) b1      conditions data base address
;       defw 0000            ;(ix+21) clen    conditions data length
;       defb 00              ;(ix+23)         (unused)
;       defw 0000            ;(ix+24) jpos    end of data for current n$() row
;       defb 00              ;(ix+26)         (unused)
;       defw 0000            ;(ix+27) goto    GOTO line number on return to BASIC
;       defb 00              ;(ix+29) klen    quantity keys list length
;       defb 00              ;(ix+30) kkey    quantity keys list key
;       defb 0000            ;(ix+31) kval    quantity keys list value
;       defb 00              ;(ix+33) crow    number of rows in c$()
;       defb 00              ;(ix+35) i   (23)current c$() index
;       defb 00              ;(ix+37) cdec(26)decrementing counter of c$() rows processed
;       defs 16              ;(ix+39) klst(35)quantity keys list

;++++++++++++++++++++++++++++++


;PRINT USR ms,string_expression,max_length,msub;

;expand string_expression and print up to max_length characters
SUBSTR: call FPTOA
        push af
        call FETCHS
        pop  af
        ex   de,hl           ;HL=string pointer
        ld   b,a             ;B =max_length
        res  0,(iy+1)        ;FLAGS "leading space"
SSNEXT1:ld   a,(hl)
        inc  hl
        push hl
        cp   $a5
        jr   nc,TOKEN
        cp   $20
        jr   nc,PRCHAR
        inc  b               ;don't count print controls
PRCHAR: rst  $10             ;print a character
        pop  hl              ;HL=string pointer
        djnz SSNEXT1         ;printed max_length?
        ret

TOKEN:  push af              ;expand this token
        sub  $a5
        ld   de,$0095
        call $0c41           ;PO_SEARCH
        pop  af              ;A =token code
        push af
        cp   197             ;OR
        jr   z,BIT0
        cp   198             ;AND
        jr   z,BIT0
        cp   202             ;LINE
        jr   c,SSNEXT2
BIT0:   bit  0,(iy+1)
        jr   nz,SSNEXT2
        ld   a,$20           ;print a leading space
        rst  $10
        dec  b
        jr   z,SSEND2        ;printed max_length?
SSNEXT2:ld   a,(de)
        push af
        and  $7f             ;reset bit 7
        rst  $10             ;print a character
        pop  af              ;A =last token char printed
        inc  de
        dec  b
        jr   z,SSEND2        ;printed max_length?
        bit  7,a
        jr   z,SSNEXT2       ;end of token?
        res  0,(iy+1)        ;FLAGS "leading space"
        pop  af              ;A =token code
        cp   168             ;FN
        jr   c,NOSPACE       ;if RND,INKEY$,PI
        cp   202             ;LINE
        jr   nc,DOSPACE      ;if token > "<>"
        cp   199             ;<=
        jr   nc,NOSPACE      ;if token <=,>=,<>
DOSPACE:ld   a,$20           ;print a trailing space
        rst  $10
        set  0,(iy+1)        ;FLAGS "no leading space"
        dec  b
        jr   z,SSEND1        ;printed max_length?
NOSPACE:pop  hl              ;HL=string pointer
        jr   SSNEXT1

SSEND2: pop  af
SSEND1: pop  hl
        ret

;++++++++++++++++++++++++++++++

;PRINT USR ms,loc,"r$()","g$()",mrg; => let r$()=g$(loc)
;PRINT USR ms,loc,"g$()","r$()",mrg; => let g$(loc)=r$()
;Transfer a variable length row from g$() to fixed length r$()
;and vice versa. Returns an empty record to r$() if an entry for
;loc is not found in g$().
SWAPRG: call GET_A           ;***
        push hl           ;1>
        call GET_A           ;***
        pop  de           ;1<
        ld   a,(hl)          ;A =destination array code
        and  %11111
        cp   18              ;destination=r$()?
        jr   z,DEST_R
        ex   de,hl
DEST_R: ld   bc,6
        add  hl,bc
        ld   (MCWORK+0),hl   ;r$() address
        ex   de,hl
        add  hl,bc
        ld   (MCWORK+2),hl   ;g$() address
        ld   (ix+4),a        ;A =destination array code
        push hl           ;1>
        call FPTOBC          ;BC=location#
        pop  hl           ;1<;HL=g$() address

SWAP_1: ld   (MCWORK+5),hl   ;=next g$() record address
        ld   e,(hl)
        ld   a,e
        inc  hl
        ld   d,(hl)          ;DE=g$() location#
        add  a,d
        jr   nz,SWAP_2

        ld   a,(ix+4)        ;no g$() row for this location#, so ...
        sub  18              ;destination=r$()?
        jr   nz,DESNOR       ;jump if not so

CLEARR: ld   hl,(MCWORK+0)   ;HL=r$() address
        ld   (hl),a          ;clear r$() & return
        ld   d,h
        ld   e,l
        inc  de
        ld   bc,127
        ldir
        ret

DESNOR: ld   (hl),b
        dec  hl
        ld   (hl),c          ;enter new loc# in g$()
        ld   d,b
        ld   e,c
        inc  hl

SWAP_2: ex   de,hl
        and  a
        sbc  hl,bc           ;(current location)-(requested location)
        ex   de,hl
        push af

        inc  hl
        ld   a,(hl)          ;=required count (total)
        inc  hl
        inc  hl
        add  a,(hl)          ;+set      count (total)
        inc  hl
        inc  hl
        add  a,(hl)          ;+reset    count (total)
        sla  a               ;(entry counts)*2 = (g$() detail length)
        inc  hl
        inc  hl
        ld   d,0
        ld   e,a             ;DE=g$() detail length
        pop  af              ;F =flags for "sbc  hl,bc"
        jr   z,SWAP_3
        add  hl,de           ;current g$() loc# doesn't match
        jr   SWAP_1          ;back to check next g$() record

SWAP_3: ld   a,(ix+4)        ;matching record found
        sub  18              ;destination=r$()?
        jr   z,EXPG2R

        ;compact r$() record to fit g$()
        ;first, make room for a longer record if necessary
COMR2G: push bc
        ld   hl,(MCWORK+0)   ;HL=r$() address
        inc  hl
        inc  hl
        ld   a,(hl)          ;=required count (total)
        inc  hl
        inc  hl
        add  a,(hl)          ;=set      count (total)
        inc  hl
        inc  hl
        add  a,(hl)          ;=reset    count (total)
        sla  a               ;(entry count)*2 = (r$() entry length)
        sub  e               ;A =(new length)-(old length)
        jr   z,ROOMOK
        ld   hl,(MCWORK+2)   ;HL=g$() address
        ld   de,2499
        add  hl,de
        ex   de,hl           ;DE=end of g$()
        ld   h,d
        ld   l,e
        ld   b,0
        ld   c,a             ;BC=extra space needed
        sbc  hl,bc           ;HL=start point for move
        push hl           ;1>
        ld   bc,(MCWORK+5)   ;BC=start of g$(loc) record
        sbc  hl,bc
        inc  hl
        ld   b,h
        ld   c,l             ;BC=bytes to move
        pop  hl           ;1<
        lddr                 ;make room for a longer record
ROOMOK: exx
        pop  bc              ;BC=location#
        ld   de,(MCWORK+5)   ;DE=start of g$(loc) record
        ld   hl,(MCWORK+0)   ;HL=start of r$() row
        ld   (hl),c
        inc  hl
        ld   (hl),b
        dec  hl
        jr   EXPCOM

        ;expand g$(loc) record to fit r$()
EXPG2R: call CLEARR          ;*** clear r$()
        exx
        ld   hl,(MCWORK+5)   ;HL=start of g$(loc) record
        ld   de,(MCWORK+0)   ;DE=start of r$() row

        ;continuation point for both options
EXPCOM: push hl           ;1>
        ld   bc,8
        ldir                 ;copy header g$()->r$()
        exx
        pop  hl           ;1<;HL=start of source row
        inc  hl
        inc  hl
        ld   b,3
        ld   d,0
EXPBAK: ld   e,(hl)          ;E =entry count  (0-20)
        sla  e               ;DE=entry length (0-40)
        push de           ;1>
        exx
        pop  bc           ;1<
        ld   a,40
        sub  c
        cp   40
        jr   z,EXPNON        ;empty entry; skip to next
        ldir
EXPNON: ld   c,a             ;BC=40-(length copied)
        ld   a,18
        cp   (ix+4)
        jr   nz,DEST_G       ;if destination g$()
        ex   de,hl           ;if destination r$()
        add  hl,bc           ;skip over unused part of r$() entry
        ex   de,hl
        jr   EXPEXX
DEST_G: add  hl,bc           ;skip over unused part of r$() entry
EXPEXX: exx
        inc  hl
        inc  hl
        djnz EXPBAK
        ret

;       org  MCWORK
;       defw 0000            ;(ix+00) r$() row address
;       defw 0000            ;(ix+02) g$() row address
;       defb 00              ;(ix+04) destination array
;       defw 0000            ;(ix+05) g$() record address

;++++++++++++++++++++++++++++++

LABELS0:
        defw ADJUST
        defw STACK
        defw PUT_T
        defw GET_T
        defw MCMOVE
        defw MCUPDT
        defw FILL
        defw COPYAB
        defw MCTRAP

LABELS1:
        defw SORT
        defw MCOND1
        defw MCOND2
        defw SUBSTR
        defw SWAPRG


;++++++++++++++++++++++++++++++