
;Shared Machine Code Routines
;============================

;Note that ORG addresses are for illustrative purposes, as actual locations
;vary depending upon the length of each program, each machine code block
;being positioned between the end of BASIC and the start of VARS.


LOCATE: equ  25000           ;CLEAR LOCATE-1
STACKP: equ  23670           ;=SEED =where SP is saved
        org  LOCATE

;adjust any absolute addresses
ADJUST: push bc
        pop  hl              ;HL=actual runtime location of ADJUST
        ld   de,LOCATE       ;DE=assembled      location of ADJUST
        and  a
        sbc  hl,de
        push hl
        pop  bc              ;BC=runtime displacement from LOCATE
        ret  z               ;return if no displacement

PUT_SP: di
        ld   (STACKP),sp     ;save stack pointer
        ld   hl,CHANGE
        add  hl,bc           ;HL=CHANGE+displacement
        ld   sp,hl           ;point SP to list of changes
ADLOOP: pop  hl
        ld   a,h
        or   l
        jr   z,GET_SP

        add  hl,bc           ;HL=location+displacement
        ex   de,hl
        pop  hl
        add  hl,bc           ;HL=content+displacement
        ex   de,hl           ;HL=address of location to change
        ld   (hl),e          ;DE=new content for that location
        inc  hl
        ld   (hl),d
        jr   ADLOOP

GET_SP: ld   sp,(STACKP)     ;restore stack pointer
        ei
        ret

        ;address relocation table
        ;terminated by 0000
        ;part 0: shared routines
        ;ld   hl,$nnnn assembles as 2A nn nn rather than ED 6B nn nn
CHANGE: defw PUT_T+1,SET_W,   GET_T+1,SET_W,    SET_W+1,GET_A, GET_A+4,FIND
        defw GET_A8+1,GET_A,  MCUPDT+1,GET_A8
        defw TO_DO2+1,CHKNEG, TO_DO2+8,CHKNEG,  TO_DO2+15,CHKNEG
        defw MCINIT+1,GET_A8
        defw LABELS0,ADJUST,  LABELS0+2,STACK,  LABELS0+4,PUT_T,  LABELS0+6,GET_T
        defw LABELS0+8,MCMOVE,LABELS0+10,MCUPDT,LABELS0+12,MCINIT,LABELS0+14,COPYAB,LABELS0+16,MCTRAP

         ;part 2: AGES-2 routines
        defw INITY0+1,GET_A8,  INDEX30+5,GET_A8, INDEX30+9,GET_A8
        defw INDEX32+1,FREESP, INDEX32+5,FREESP, LDTEXT+1,TEXT,    LDTEXT+7,OUTVAL, LDTEXT+13,OUTVAL
        defw SPACEN+1,OUTVAL,  SPACEN+7,OUTVAL,  NEXTA+1,ADDLOC,   NEXTJP+1,FORA
        defw ADD_JP+1,NEXTB,   CALLC2+1,CHUNK2,  CALLI2+1,INDEX2,  NEXTB+7,ADDLOC,   NEXTB+16,ADDLOC
        defw CHECKY+1,GET_A8,  LDTXNL+1,TEXTNL,  BASRET-2,FORI
        defw LABELS1,INDEX2,   LABELS1+2, INITY0,LABELS1+4, INDEX30,LABELS1+6, CHECKL

        ;end marker
        defw $0000

CLOWER: equ  $0d6e           ;CLS-LOWER clear lower screen
CHOPEN: equ  $1601           ;CHAN-OPEN open channel A
OUTNUM: equ  $1a1b           ;OUT-NUM-1 print integer BC 0-9999
PRINTS: equ  $203c           ;PR-STRING print a string length BC from DE
PLOT:   equ  $22dc           ;PLOT      plot a point
DRAW:   equ  $2477           ;LINE-DRAW draw a line
SCANNG: equ  $24fb           ;the SCANNING subroutine
STORES: equ  $2ab6           ;to calculator stack: AEDCB
FETCHS: equ  $2bf1           ;from calculator stack: AEDCB
STAKA:  equ  $2d28           ;to calculator stack: A
STAKBC: equ  $2d2b           ;to calculator stack: BC
FPTOBC: equ  $2da2           ;from calculator stack: BC
FPTOA:  equ  $2dd5           ;from calculator stack: A

CH_ADD: equ  23645
MCWORK: equ  65368

;stack function parameters passed from BASIC by:
;PRINT USR mcstack,arg1[,arg2...,argn],addr;
STACK:  ld   hl,14
        add  hl,sp
        ld   sp,hl           ;undo USR
        ld   ix,0
GETARG: rst  $20
        call SCANNG
        inc  ix
        rst  $18
        cp   ","
        jr   z,GETARG
        call FPTOBC          ;BC=addr
        ld   h,b
        ld   l,c
        push ix
        pop  bc
        dec  bc
        ld   a,c             ;A =number of parameters
        ld   ix,MCWORK
        jp   (hl)            ;JP addr


;find a string array; call with DE=address of array name
FIND:   ld   bc,(CH_ADD)
        push bc
        ld   (CH_ADD),de
        call $28b2            ;LOOK-VARS
        pop  bc
        ld   (CH_ADD),bc
        ret

;	PRINT USR mcstack,"t$()",t;
;     t$()=array name,
;     t=address of machine code routine
;      =mcputt : put a copy of header array named in t$ into workspace
;      =mcgett : get a copy of header array named in t$ from workspace

;put a copy of header array t$ into workspace
PUT_T:  call SET_W            ;***
        ldir
        ret
;get a copy of header array t$ from workspace
GET_T:  call SET_W            ;***
        ex   de,hl
        ldir
        ret

;set header array & workspace addresses
SET_W:  call GET_A           ;***HL=array address
        ld   de,65536-168    ;DE=workspace address
        ld   bc,54           ;BC=header array size (=48 + 6 byte header)
        ret

;get array address
GET_A:  call FETCHS          ;DE=array name address
        call FIND            ;***HL=array address
        ret
GET_A8: call GET_A          ;***HL=array address
        ld   bc,8
        add  hl,bc          ;HL=address of 1st row in a 2D array
        ret

;get a +ve/-ve integer off the calculator stack
CHKNEG: call FPTOBC
        ret  z               ;return if not -ve
        ld   hl,0000
        and  a
        sbc  hl,bc           ;negate BC
        ld   b,h
        ld   c,l
        ret


;	 PRINT USR mcstack,l,n,o,mcmove;
;     l=length
;     n=new address (higher/lower)
;     o=old address

;move a chunk of the exits data
MCMOVE: call FPTOBC          ;BC=old address
        push bc
        call FPTOBC          ;BC=new address
        push bc
        call FPTOBC          ;BC=length
        pop  de              ;DE=new address
        and  a               ;reset Carry flag
        pop  hl              ;HL=old address
CHUNK2: push hl
        sbc  hl,de           ;old address - new address
        pop  hl
        jr   c,HIGHER
LOWER:  ldir                 ;old address > new address
        ret
HIGHER: add  hl,bc           ;old address < new address
        dec  hl              ;HL=end of old chunk
        push hl
        ex   de,hl
        add  hl,bc
        dec  hl
        ex   de,hl           ;DE=end of new chunk
        pop  hl
        lddr
        ret


;	 PRINT USR mcstack,d,a,s,l,r,f,i$,mcupdt;

;     d=row displacement*rowlen =byte displacement(+/-)
;     a=adjustment(+/-)
;     s=row step*rowlen         =byte step(+/-rowlen)
;     l=stop  row #
;     r=row length
;     f=start row #
;     i$=array name

;for rows f through l of array named in i$
;adjust address pointers by a if greater than p
;shift new row data by (+/-)d bytes from old row location
;step through old row data by (+/-)s bytes

;update a range of index array rows
MCUPDT: call GET_A8          ;***
        push hl          ;a1:>HL=first row address
        call FPTOBC
        push bc          ;a2:>BC=start row #
        call FPTOBC          ;BC=row length
        pop  de          ;a2:<DE=start row #
        pop  hl          ;a1:<HL=first row address
        push bc          ;b1:>BC=row length
        push hl          ;b2:>HL=first row address
        dec  de              ;DE=start row # - 1
        ld   b,c             ;B =row length
        ld   hl,0000
ADD_UP: add  hl,de           ;(start row#-1)*(row length)
        djnz ADD_UP          ;HL=start row offset
        inc  de              ;DE=start row #
        pop  bc          ;b2:<BC=first row address
        add  hl,bc
        push hl          ;c1:>HL=start row address
        push de          ;c2:>DE=start row #
        call FPTOBC          ;BC=stop row #
        pop  hl          ;c2:<HL=start row #
        xor  a
        sbc  hl,bc
        jr   nc,TO_DO        ;jump if start row # >= stop row #
        ld   d,a             ;sign -ve
        ld   e,a             ;DE=0000
        ex   de,hl
        and  a
        sbc  hl,de           ;reverse sign to +ve
TO_DO:  inc  hl
        pop  de          ;c1:<DE=start row address
        push hl          ;d1:>HL=number of rows to do
        exx
        pop  bc          ;d1:<BC'=number of rows to do
        exx
        push de          ;b2:>DE=start row address
TO_DO2: call CHKNEG          ;***BC=row step*rowlen =byte step(+/-rowlen)
        ld   (MCWORK+75),bc
        call CHKNEG          ;***BC=adjustment(+/-)
        ld   (MCWORK+77),bc
        call CHKNEG          ;***BC=row displacement*rowlen(+/-)
        ld   (MCWORK+79),bc
        pop  hl          ;b2:<HL=start row address
NEXT:   ld   e,(hl)          ;HL=old row address
        inc  hl
        ld   d,(hl)          ;DE=old pointer
        ld   a,d
        or   e
        jr   z,NOADD         ;don't update if old pointer is zero
TO_DO6: ld   bc,(MCWORK+77)  ;BC=adjustment
        ex   de,hl           ;HL=old pointer
        add  hl,bc           ;HL=old pointer + adjustment
        ex   de,hl           ;DE=new pointer
NOADD:  ld   (hl),d
        dec  hl
        ld   (hl),e          ;put new pointer at old address
        ex   de,hl           ;DE=old row address
TO_DO7: ld   hl,(MCWORK+79)  ;HL=displacement
        add  hl,de           ;HL=old row address + displacement (may be 0)
        ex   de,hl           ;HL=old row address; DE=new row address
        pop  bc          ;b1:<BC=row length
        push bc          ;b1:>BC=row length
        push hl          ;b2:>HL=old row address
        ldir                 ;copy updated row to new address
        pop  hl          ;b2:<HL=old row address
TO_DO5: ld   bc,(MCWORK+75)  ;BC=step
        add  hl,bc           ;HL=old address + step
        exx
        dec  bc              ;decrement row counter
        ld   a,b
        or   c
        exx
        jr   nz,NEXT
        pop  bc          ;b1:<BC=row length
        ret


;PRINT USR mcstack,l,x,n$,mcinitn;
MCINIT: call GET_A8         ;***
        push hl             ;HL=address of 1st row in array
        call FPTOBC         ;BC=number of rows
        push bc
        call FPTOA          ;A =row length
        ld   b,a
        ld   c,a
        pop  de             ;DE=number of rows
        pop  hl
        push hl             ;save original array pointer
REPS:   ld   (hl),$00       ;initialise 1st row
        inc  hl
        djnz REPS
        push hl             ;save updated array pointer
        dec  de             ;DE=(number of rows)-1 for row already done
        ld   hl,0000
        ld   b,c
REPD:   add  hl,de          ;multiply (row#-1)*rowlen
        djnz REPD
        ld   b,h
        ld   c,l            ;BC=number of bytes to copy
        pop  de             ;DE=starting destination (1+rowlen byte of array)
        pop  hl             ;HL=starting source      (1st byte of array)
        ldir                ;initialise remainder of array rows
        ret


;PRINT USR mcstack,a,b,l[,"c"],copy;
;[optionally copy string 'c' to 'a' then] copy 'l' bytes from 'a' to 'b' (may overlap)
COPYAB: cp   4
        jr   z,COPY2
        xor  a
        call STAKA      ;stack zero for an empty string
COPY2:  call FETCHS     ;get initialisation string
        push bc      ;1>;BC=string length
        push de      ;2>;DE=string address
        call FPTOBC
        pop  de      ;2<
        pop  hl      ;1<
        push bc      ;1>;BC=copy length
        push hl      ;2>
        push de      ;3>
        call FPTOBC
        pop  de      ;3<
        pop  hl      ;2<
        push bc      ;2>;BC=copy to address
        push hl      ;3>
        push de      ;4>
        call FPTOBC
        ld   d,b
        ld   e,c        ;DE=copy from address
        pop  hl      ;4<;HL=initialisation string address
        pop  bc      ;3<;BC=initialisation string length
        push de      ;3>
        ld   a,b
        or   c
        jr   z,COPY3
        ldir            ;copy initialisation string
        scf
COPY3:  pop  hl      ;3<
        pop  de      ;2<
        pop  bc      ;1<
        jr   c,COPYLO    ;jump if init string set
        push hl
        sbc  hl,de      ;from - to
        pop  hl
        jr   c,COPYHI
COPYLO: ldir            ;from > to -or- initialisation string specified
        ret
COPYHI: add  hl,bc      ;from < to
        dec  hl         ;HL=end of from chunk
        push hl
        ex   de,hl
        add  hl,bc
        dec  hl
        ex   de,hl      ;DE=end of to chunk
        pop  hl
        lddr
        ret


;ERROR TRAP - machine code segment

MCTRAP: RES  4,(IY+1)       ;patch for +3 128k mode

        LD   A,(23621)      ;save PPC line number (where error occurred)
        LD   (23615),A      ;in LIST_SP
        LD   A,(23622)
        LD   (23616),A
        LD   A,(23610)      ;save ERR_NR error code
        LD   (23560),A      ;in LAST_K
        LD   BC,9000        ;new line number to jump to
        LD   (23662),BC     ;OLDPPC - CONTINUE line number
        LD   A,$ff
        LD   (23610),A      ;ERR_NR - reset error code
        INC  A
        LD   (23664),A      ;OSPCC  - CONTINUE statement number
        LD   HL,(23641)     ;E_LINE - address of editor buffer
        CALL $1652          ;ONE-SPACE - make a space in it
        EX   DE,HL
        LD   (HL),$e8       ;insert code for "CONTINUE"
MCLOOP: CALL $1f54          ;BREAK-KEY
        JP   C,$12cf        ;MAIN-3 - jump if not BREAK
        JR   MCLOOP         ;ignore BREAK

;++++++++++++++++++++++++++++++


;update a range of y$() index array rows
INDEX2: ld   l,(ix+14)
        ld   h,(ix+15)       ;HL=array address (skipping header)
        push de          ;a1>;DE=start row #
        push hl          ;a2>
        dec  de              ;DE=start row # - 1
        ld   h,d
        ld   l,e
        add  hl,hl
        add  hl,de           ;multiply (row#-1)*length (3)
        ex   de,hl           ;DE=start row offset
        pop  hl          ;a2<;HL=array address
        add  hl,de           ;HL=start row address
        pop  de          ;a1<;DE=start row #
        push hl          ;a1>;save start row address
        ld   c,(ix+16)
        ld   b,(ix+17)       ;BC=stop row #
        ld   h,d
        ld   l,e             ;HL=start row #
        and  a
        sbc  hl,bc
        jr   nc,TO_DO22      ;jump if start row # >= stop row #
        ld   h,b
        ld   l,c             ;HL=stop row #
        and  a
        sbc  hl,de
TO_DO22:inc  hl              ;HL=number of rows to do
        push hl          ;a2>;save number of rows to do
        pop  bc          ;a2<;BC=number of rows to do
        pop  hl          ;a1<;get address of start row
NEXT2:  push hl          ;a1>
        ld   e,(hl)
        inc  hl
        ld   d,(hl)          ;DE=old pointer
        inc  hl
        ld   a,(hl)          ;A =entry length
        inc  de
        inc  de
        inc  de              ;DE=old pointer + adjustment (3)
        pop  hl          ;a1<;get address of row updated
        push hl          ;a1>
        ld   (hl),e
        inc  hl
        ld   (hl),d          ;put new pointer at new address
        inc  hl
        ld   (hl),a          ;put entry length at new address
        pop  hl          ;a1<;get address of row updated
        inc  hl
        inc  hl
        inc  hl           ;HL=old address + step (6)
        dec  bc              ;decrement row counter
        ld   a,b
        or   c
        jr   nz,NEXT2
        ret

;------


;PRINT USR mcstack,edat,b1,irow,"y$()",mcinity;

INITY0: call GET_A8           ;***
        push hl          ;a1>;HL=y$() address (skipping header)
        call FPTOBC          ;BC=number of rows to do
        PUSH BC          ;a2>
        call FPTOBC          ;BC=b1
        PUSH BC          ;a3>
        call FPTOBC          ;BC=edat (b1 size)
        pop  hl          ;a3<;HL=b1
        ld   (hl),0          ;initialise 1st exit item(b1)
        inc  hl
        ld   (hl),$ff
        inc  hl
        ld   (hl),0
        inc  hl
        dec  bc              ;BC=edat (b1 size)
        dec  bc              ;deduct 3 for item already done
        dec  bc
        ex   de,hl           ;DE=starting destination (b1+3)
        ld   h,d
        ld   l,e
        dec  hl
        dec  hl
        dec  hl              ;HL=starting source      (b1)
        PUSH HL          ;a3>
        ldir                 ;initialise remainder of exits data
        pop  de          ;a3<;DE=start of exits data  (b1)
        pop  bc          ;a2<;BC=counter
        pop  hl          ;a1<;HL=y$() address
NEXTJ:  ld   (hl),e
        inc  hl
        ld   (hl),d          ;set y$() initial entry address
        inc  hl
        ld   (hl),3          ;set y$() initial entry length
        inc  hl
        inc  de
        inc  de
        inc  de              ;DE=b1 address of next exits data item
        dec  bc              ;decrement row counter
        ld   a,b
        or   c
        jr   nz,NEXTJ
        ret


;    PRINT USR mcstack,emax,blen,b1,elen,"y$()","i$()",enum,mcupdt3;
;    LET blen=FN p(23675): LET free=FN p(23677)
;    IF NOT free THEN
;         LET e$="NO ROOM": GO SUB 1900
;    RETURN

TEXT:   defm "LocMaxSizSpc"

INDEX30:call FPTOBC
        push bc              ;=enum
        call GET_A8          ;***
        push hl              ;=i$() base address
        call GET_A8          ;***
        push hl              ;=y$() base address
        call FPTOBC
        push bc              ;=elen
        call FPTOBC
        push bc              ;=b1
        call FPTOBC
        push bc              ;=blen
        call FPTOBC
        push bc              ;=emax

        ld   bc,MCWORK
        push bc
        pop  ix              ;=mcwork
        ld   (MCWORK+8),bc
        pop  bc              ;=emax
        ld   (MCWORK+28),bc
        pop  bc              ;=blen
        ld   (MCWORK+18),bc
        pop  bc              ;=b1
        ld   (ix+26),c
        ld   (ix+27),b
INDEX32:call FREESP          ;***=mcwork-b1
        pop  bc              ;=elen
        call FREESP          ;***=mcwork-b1-elen
        pop  hl              ;=y$() base address
        ld   (MCWORK+14),hl
        pop  hl              ;=i$() base address
        ld   (MCWORK+12),hl
        pop  bc              ;=enum
        ld   (MCWORK+16),bc
        push bc              ;start loop counter=enum

        ld   de,0
FORA:   inc  de              ;DE=a
        ld   (MCWORK+20),de  ;save 'a' for ADDLOC
        push de              ;save 'a' for end of loop

;    PRINT #0;AT k0,29;a:
        xor  a               ;stream #0
        call CHOPEN
        ld   a,22            ;AT
        rst  $10
        xor  a               ;0
        rst  $10
        ld   a,29            ;29
        rst  $10
        pop  bc              ;BC=a
        push bc
        call OUTNUM

;    PRINT AT k0,k0; "Loc"; INVERSE 1;enum;
;         INVERSE k0;"Max"; INVERSE 1;emax;
;         INVERSE k0;"Siz"; INVERSE 1;blen+(elen AND NOT blen);
;         INVERSE k0;"Spc"; INVERSE 1;free-(free AND free<k0);
;         INVERSE k0;TAB 26;"AGES-2"
        ld   a,2             ;stream #2
        call CHOPEN
        ld   a,22            ;AT
        rst  $10
        xor  a               ;0
        rst  $10
        xor  a               ;0
        rst  $10
LDTEXT: ld   de,TEXT         ;***
        ld   hl,(MCWORK+16)  ;HL=enum
        call OUTVAL          ;***
        ld   hl,(MCWORK+28)  ;HL=emax
        call OUTVAL          ;***
;free=mcwork-b1-elen-blen
        and  a
        ld   hl,(MCWORK+8)   ;HL=mcwork-b1-elen
        ld   bc,(MCWORK+18)  ;BC=blen
        sbc  hl,bc           ;HL=free=mcwork-b1-elen-blen
        ld   (MCWORK+10),hl  ;=free
        ld   l,c
        ld   h,b             ;HL=blen
SPACEN: call OUTVAL          ;***
        ld   hl,(MCWORK+10)  ;HL=free
        call OUTVAL          ;***
        ld   a,23            ;TAB
        rst  $10
        ld   a,26            ;26
        rst  $10
        xor  a               ;0
        rst  $10

;process i$(a)
        ld   hl,(MCWORK+12)  ;HL=i$(a) address
        ld   c,(hl)
        ld   (ix+22),c
        inc  hl
        ld   b,(hl)
        ld   (ix+23),b       ;BC=ibeg
        inc  hl
        ld   e,(hl)
        ld   d,0             ;DE=ilen
        push de
        inc  hl
        inc  hl              ;+++
        inc  hl              ;+++
        inc  hl              ;+++
        ld   (MCWORK+12),hl  ;=next i$(a) address
        ex   de,hl           ;HL=ilen
        add  hl,bc           ;HL=iend
        ld   (MCWORK+24),hl  ;=iend
        pop  de              ;DE=ilen
        ld   hl,(MCWORK+10)  ;HL=free
        and  a
        sbc  hl,de
        jr   nc,NEXTA
NOROOM: pop  de
        pop  bc
RETBAS: ld   a,(ix+18)
        ld   (23675),a
        ld   a,(ix+19)
        ld   (23676),a       ;UDG=blen
        ld   a,(ix+10)
        ld   (23677),a
        ld   a,(ix+11)
        ld   (23678),a       ;COORDS=free
        ret

NEXTA:  call ADDLOC          ;***
        pop  de              ;DE=a
        pop  bc              ;BC=loop counter
        dec  bc
        ld   a,c
        or   b
        jr   z,RETBAS
        push bc
NEXTJP: jp   FORA            ;***

;store mcwork[[-b1]-elen]
FREESP: ld   hl,(MCWORK+8)
        and  a
        sbc  hl,bc
        ld   (MCWORK+8),hl
        ret

;print text pointed to by DE followed by value in HL
OUTVAL: push hl
        ld   bc,3            ;BC=length of this text
        ld   l,e
        ld   h,d             ;HL=address of this text
        add  hl,bc           ;HL=address of next text
        push hl
        call PRINTS
        ld   a,20            ;INVERSE
        rst  $10
        ld   a,1             ;1
        rst  $10
        pop  de              ;DE=address of next text
        pop  bc              ;BC=value to print
        push de
        call OUTNUM
        ld   a,20            ;INVERSE
        rst  $10
        xor  a               ;0
        rst  $10
        pop  de              ;DE=address of next text
        ret

ADDLOC: ld   hl,(MCWORK+22)  ;HL=ibeg=exit address (b)
        ld   e,(hl)
        inc  hl
        ld   d,(hl)          ;DE=location number (y$() index (n))
        ld   (MCWORK+0),de
        inc  hl
        inc  hl              ;+++
        ld   (MCWORK+22),hl  ;store next exit address
        ld   a,e
        or   d
ADD_JP: jp   z,NEXTB         ;***
        dec  de
        ld   h,d
        ld   l,e
        add  hl,hl
        add  hl,de           ;HL=(row#-1)*3=byte offset of y$() row
        ld   de,(MCWORK+14)  ;DE=y$() array base address
        add  hl,de           ;HL=y$(n) row's address
        ld   (MCWORK+2),hl   ;PUSH hl
        ld   e,(hl)
        inc  hl
        ld   d,(hl)          ;DE=y$(n) row's entry data address (jbeg)
        ld   (MCWORK+4),de   ;PUSH de
        inc  hl
        ld   a,(hl)          ;A =y$(n) row's entry data length  (jlen)
        ld   (ix+6),a
        ex   de,hl           ;HL=y$(n) row's entry data address (jbeg)
        inc  hl
        ld   a,$ff
        cp   (hl)            ;check if 1st entry is unused (hi=$ff)
        jr   z,UNUSED
        ld   a,2             ;+++
UNUSED: inc  a               ;A =length modifier, 00 or 03 (ldif)
        or   a
        jr   z,NOMOVE        ;IF NOT ldif ...
        ld   hl,(MCWORK+0)   ;HL=location number (y$() index (n))
        ld   de,(MCWORK+16)  ;DE=total locations (enum)
        and  a
        sbc  hl,de
        jr   nc,NOMOVE       ;... OR n>=enum
        ld   hl,(MCWORK+2)   ;HL=y$(n) row's address
        inc  hl
        inc  hl
        inc  hl           ;HL=y$(n+1) row's address
        ld   e,(hl)
        inc  hl
        ld   d,(hl)          ;DE=y$(n+1) row's entry data address (jpos)
        ld   h,d
        ld   l,e             ;HL=y$(n+1) row's old entry data address
        inc  de
        inc  de              ;+++
        inc  de              ;DE=y$(n+1) row's new entry data address
        PUSH hl
        ld   hl,(MCWORK+26)  ;HL=b1
        ld   bc,(MCWORK+18)  ;BC=blen
        add  hl,bc           ;HL=b1+blen
        sbc  hl,de           ;HL=b1+blen-(jpos+3)
        inc  hl
        inc  hl              ;+++
        inc  hl              ;HL=b1+blen-jpos
        ld   b,h
        ld   c,l             ;BC=number of bytes to copy
        POP  hl              ;HL=source, DE=destination
        PUSH af
CALLC2: call CHUNK2          ;***
        ld   de,(MCWORK+0)   ;DE=location number (y$() index (n))
        inc  de              ;DE=location number (n+1)
CALLI2: call INDEX2          ;***
        POP  af
NOMOVE: ld   b,(ix+6)        ;B =jlen
        ld   c,a             ;C =ldif
        ld   de,(MCWORK+4)   ;DE=jbeg
        ld   hl,(MCWORK+2)   ;HL=y$(n) row's address
        PUSH bc
        inc  hl
        inc  hl              ;HL=y$(n) row's entry length address
        add  a,(hl)
        ld   (hl),a          ;update y$(n) row's entry length
        ld   a,b             ;A =jlen
        add  a,c             ;A =jlen+ldif =nlen
        ld   c,a
        ld   b,0             ;BC=nlen
        ex   de,hl           ;HL=jbeg
        add  hl,bc           ;HL=jbeg+nlen
        dec  hl
        dec  hl
        ld   a,(ix+21)
        ld   (hl),a          ;POKE jbeg+nlen-2,FN h(a)
        dec  hl
        ld   a,(ix+20)
        ld   (hl),a          ;POKE jbeg+nlen-3,FN g(a)
        ld   hl,(MCWORK+18)  ;HL=y$() data length (blen)
        POP  bc              ;BC=jlen,ldif
        ld   b,0             ;BC=ldif
        add  hl,bc           ;LET blen=blen+ldif
        ld   (MCWORK+18),hl  ;=blen
NEXTB:  ld   a,(ix+22)
        cp   (ix+24)
        jp   nz,ADDLOC       ;***
        ld   a,(ix+23)
        cp   (ix+25)
        jp   nz,ADDLOC       ;***
        ret

;++++++++++++++++++++

;    PRINT USR mcstack,"y$()",bmax,mccheckl;

TEXTNL: defm "Missing link to<"
        defb $00,$00,$00

CHECKL: xor  a               ;stream #0
        call CHOPEN
        call FPTOBC
        push bc          ;a1>
        pop  de          ;a1<
        ld   ix,MCWORK
        ld   (MCWORK+28),de  ;DE=bmax
        dec  de
        push de          ;a1>

CHECKY: call GET_A8          ;***
        ld   (MCWORK+14),hl  ;=y$() base address
        ld   a,1
        ld   (ix+20),a
        xor  a
        ld   (ix+21),a       ;=y$(i) index

FORI:   inc  hl
        inc  hl
        inc  hl
        push hl          ;a2>;save pointer for end of loop
        push hl          ;a3>;save pointer while printing progress counter

;    PRINT #0;AT k0,29;i:
        ld   a,22            ;AT
        rst  $10
        xor  a               ;0
        rst  $10
        ld   a,29            ;29
        rst  $10
        ld   bc,(MCWORK+20)  ;BC=i
        inc  bc              ;BC=i+1
        ld   (MCWORK+20),bc
        call OUTNUM

;process y$(i)
        pop  hl          ;a3<;HL=y$(i) address
        ld   e,(hl)
        ld   (ix+22),e
        inc  hl
        ld   d,(hl)
        ld   (ix+23),d       ;DE=entry start
        inc  hl
        ld   c,(hl)
        ld   b,0             ;BC=entry length

FORE:   push bc          ;a3>
        ld   a,(de)
        ld   l,a
        inc  de
        ld   a,(de)
        ld   h,a             ;HL=entry number
        inc  de
        inc  de
        ld   a,l
        or   h
        jr   z,NOLINK        ;FN p(e)=k0
        and  a
        ld   bc,1000
        sbc  hl,bc
        jr   nc,NOLINK       ;FN p(e)>999

NEXTE:  pop  bc          ;a3<;BC=inner loop counter
        dec  c
        dec  c
        dec  c
        jr   nz,FORE
        jr   NEXTI

NOLINK: push de          ;a4>
        ex   de,hl           ;HL=entry address+3
        dec  hl
        dec  hl
        ld   (hl),0
        dec  hl
        ld   (hl),0          ;reset entry number
;PRINT "Missing link to<";i;">"
        ld   a,2             ;stream #2
        call CHOPEN
LDTXNL: ld   de,TEXTNL       ;***
        ld   bc,16
        call PRINTS
        ld   bc,(MCWORK+20)  ;BC=i
        call OUTNUM
        ld   a,">"
        rst  $10
        ld   a,$0d           ;ENTER
        rst  $10
        xor  a               ;stream #0
        call CHOPEN
        pop  de          ;a4<;DE=next entry address
        jr   NEXTE

NEXTI:  pop  hl          ;a2<;HL=y$()pointer
        pop  bc          ;a1<;BC=outer loop counter
        dec  bc
        ld   a,c
        or   b
        jr   z,BASRET
        push bc          ;a1>
        jp   FORI            ;***
BASRET: ld   a,2             ;stream #2
        call CHOPEN
        ret


;        org  MCWORK
;        defb 0,0           ;location number (y$() index (n))
;        defb 0,0           ;y$(n) row address
;        defb 0,0           ;y$(n) row entry data address (jbeg)
;        defb 0             ;y$(n) row entry data length  (jlen)
;        defb 0
;        defb 0,0           ;mcwork-b1-elen          (IX+8,9)
;        defb 0,0           ;free                    (IX+10,11)
;        defb 0,0           ;i$() array base address (IX+12,13)
;        defb 0,0           ;y$() array base address (IX+14,15)
;        defb 0,0           ;total locations,        (IX+16,17)= enum
;        defb 0,0           ;y$() data length,       (IX+18,19)= blen
;        defb 0,0  [INDEX3] ;i$() index #,           (IX+20,21)= a
;                  [CHECKL] ;y$() index #,           (IX+20,21)= i
;        defb 0,0  [INDEX3] ;i$() 1st  exit address  (IX+22,23)= ibeg
;                  [CHECKL] ;y$() 1st  exit address  (IX+22,23)= start
;        defb 0,0           ;i$() last exit address  (IX+24,25)= iend
;        defb 0,0           ;y$() data base address  (IX+26,27)= b1
;        defb 0,0  [INDEX3] ;emax                    (IX+28,29)
;                  [CHECKL] ;bmax                    (IX+28,29)

;--------

LABELS0:
        defw ADJUST
        defw STACK
        defw PUT_T
        defw GET_T
        defw MCMOVE
        defw MCUPDT
        defw MCINIT
        defw COPYAB
        defw MCTRAP

LABELS1:
        defw INDEX2
        defw INITY0
        defw INDEX30
        defw CHECKL