
;Shared Machine Code Routines
;============================

;Note that ORG addresses are for illustrative purposes, as actual locations
;vary depending upon the length of each program, each machine code block
;being positioned between the end of BASIC and the start of VARS.


LOCATE: equ  25000           ;CLEAR LOCATE-1
STACKP: equ  23670           ;=SEED =where SP is saved
        org  LOCATE

;adjust any absolute addresses
ADJUST: push bc
        pop  hl              ;HL=actual runtime location of ADJUST
        ld   de,LOCATE       ;DE=assembled      location of ADJUST
        and  a
        sbc  hl,de
        push hl
        pop  bc              ;BC=runtime displacement from LOCATE
        ret  z               ;return if no displacement

PUT_SP: di
        ld   (STACKP),sp     ;save stack pointer
        ld   hl,CHANGE
        add  hl,bc           ;HL=CHANGE+displacement
        ld   sp,hl           ;point SP to list of changes
ADLOOP: pop  hl
        ld   a,h
        or   l
        jr   z,GET_SP

        add  hl,bc           ;HL=location+displacement
        ex   de,hl
        pop  hl
        add  hl,bc           ;HL=content+displacement
        ex   de,hl           ;HL=address of location to change
        ld   (hl),e          ;DE=new content for that location
        inc  hl
        ld   (hl),d
        jr   ADLOOP

GET_SP: ld   sp,(STACKP)     ;restore stack pointer
        ei
        ret

        ;address relocation table
        ;terminated by 0000
        ;part 0: shared routines
        ;ld   hl,$nnnn assembles as 2A nn nn rather than ED 6B nn nn
CHANGE: defw PUT_T+1,SET_W,   GET_T+1,SET_W,    SET_W+1,GET_A, GET_A+4,FIND
        defw GET_A8+1,GET_A,  MCINIT+1,GET_A8
        defw LABELS0,ADJUST,  LABELS0+2,STACK,  LABELS0+4,PUT_T,  LABELS0+6,GET_T
        defw LABELS0+8,MCINIT,LABELS0+10,COPYAB,LABELS0+12,MCTRAP

         ;part 3: AGES-3 routines
        defw NXTARR+2,GET_A8, MCBACK+8,NXTARR, MCBACK+17,NXTINT
        defw G1224A+1,GO1224, G1222A+1,GO1222, G1222B+1,GO1222
        defw NEXTA+1,JPGTLE,  CHEKHS+2,GO1216, NEXTI+1,FORIEM
        defw FTEXT1+1,FCDESC, FTEXT2+1,FFOUND
        defw MCFRWD+8,NXTARR, MCFRWD+13,NXTINT,GO1244-2,GO1232, NEXTI2+1,GO1232
        defw MCCOMP+8,NXTARR, MCCOMP+13,NXTINT,NEXTI3+1,FOR1EM
        defw MCSAVE+8,NXTARR, MCSAVE+13,NXTINT,NEXTI4+1,FOR2K1
        defw READRG+1,GET_A,  READRG+5,GET_A,  EXPG2R+1,CLEARR
        defw LABELS1,TOPRAM,  LABELS1+2,MCBACK,LABELS1+4,MCFRWD,LABELS1+6,MCCOMP
        defw LABELS1+8,MCSAVE,LABELS1+10,READRG

        ;end marker
        defw $0000

CLOWER: equ  $0d6e           ;CLS-LOWER clear lower screen
CHOPEN: equ  $1601           ;CHAN-OPEN open channel A
OUTNUM: equ  $1a1b           ;OUT-NUM-1 print integer BC 0-9999
PRINTS: equ  $203c           ;PR-STRING print a string length BC from DE
PLOT:   equ  $22dc           ;PLOT      plot a point
DRAW:   equ  $2477           ;LINE-DRAW draw a line
SCANNG: equ  $24fb           ;the SCANNING subroutine
STORES: equ  $2ab6           ;to calculator stack: AEDCB
FETCHS: equ  $2bf1           ;from calculator stack: AEDCB
STAKA:  equ  $2d28           ;to calculator stack: A
STAKBC: equ  $2d2b           ;to calculator stack: BC
FPTOBC: equ  $2da2           ;from calculator stack: BC
FPTOA:  equ  $2dd5           ;from calculator stack: A

CH_ADD: equ  23645
MCWORK: equ  65368

;stack function parameters passed from BASIC by:
;PRINT USR mcstack,arg1[,arg2...,argn],addr;
STACK:  ld   hl,14
        add  hl,sp
        ld   sp,hl           ;undo USR
        ld   ix,0
GETARG: rst  $20
        call SCANNG
        inc  ix
        rst  $18
        cp   ","
        jr   z,GETARG
        call FPTOBC          ;BC=addr
        ld   h,b
        ld   l,c
        push ix
        pop  bc
        dec  bc
        ld   a,c             ;A =number of parameters
        ld   ix,MCWORK
        jp   (hl)            ;JP addr


;find a string array; call with DE=address of array name
FIND:   ld   bc,(CH_ADD)
        push bc
        ld   (CH_ADD),de
        call $28b2            ;LOOK-VARS
        pop  bc
        ld   (CH_ADD),bc
        ret

;	PRINT USR mcstack,"t$()",t;
;     t$()=array name,
;     t=address of machine code routine
;      =mcputt : put a copy of header array named in t$ into workspace
;      =mcgett : get a copy of header array named in t$ from workspace

;put a copy of header array t$ into workspace
PUT_T:  call SET_W            ;***
        ldir
        ret
;get a copy of header array t$ from workspace
GET_T:  call SET_W            ;***
        ex   de,hl
        ldir
        ret

;set header array & workspace addresses
SET_W:  call GET_A           ;***HL=array address
        ld   de,65536-168    ;DE=workspace address
        ld   bc,54           ;BC=header array size (=48 + 6 byte header)
        ret

;get array address
GET_A:  call FETCHS          ;DE=array name address
        call FIND            ;***HL=array address
        ret
GET_A8: call GET_A          ;***HL=array address
        ld   bc,8
        add  hl,bc          ;HL=address of 1st row in a 2D array
        ret

;;get a +ve/-ve integer off the calculator stack
;CHKNEG: call FPTOBC
;        ret  z               ;return if not -ve
;        ld   hl,0000
;        and  a
;        sbc  hl,bc           ;negate BC
;        ld   b,h
;        ld   c,l
;        ret


;PRINT USR mcstack,l,x,n$,mcinitn;
MCINIT: call GET_A8         ;***
        push hl             ;HL=address of 1st row in array
        call FPTOBC         ;BC=number of rows
        push bc
        call FPTOA          ;A =row length
        ld   b,a
        ld   c,a
        pop  de             ;DE=number of rows
        pop  hl
        push hl             ;save original array pointer
REPS:   ld   (hl),$00       ;initialise 1st row
        inc  hl
        djnz REPS
        push hl             ;save updated array pointer
        dec  de             ;DE=(number of rows)-1 for row already done
        ld   hl,0000
        ld   b,c
REPD:   add  hl,de          ;multiply (row#-1)*rowlen
        djnz REPD
        ld   b,h
        ld   c,l            ;BC=number of bytes to copy
        pop  de             ;DE=starting destination (1+rowlen byte of array)
        pop  hl             ;HL=starting source      (1st byte of array)
        ldir                ;initialise remainder of array rows
        ret


;PRINT USR mcstack,a,b,l,copy;
;copy 'l' bytes from 'a' to 'b' (may overlap)
COPYAB: call FPTOBC
        push bc      ;1>;BC=copy length
        call FPTOBC
        push bc      ;2>;BC=copy to address
        call FPTOBC
        ld   h,b
        ld   l,c        ;HL=copy from address
        pop  de      ;2<;DE=copy to address
        pop  bc      ;1<;BC=copy length
        push hl
        sbc  hl,de      ;from - to
        pop  hl
        jr   c,COPYHI
COPYLO: ldir            ;from > to
        ret
COPYHI: add  hl,bc      ;from < to
        dec  hl         ;HL=end of from chunk
        push hl
        ex   de,hl
        add  hl,bc
        dec  hl
        ex   de,hl      ;DE=end of to chunk
        pop  hl
        lddr
        ret


;ERROR TRAP - machine code segment

MCTRAP: RES  4,(IY+1)       ;patch for +3 128k mode
        LD   A,(23621)      ;save PPC line number (where error occurred)
        LD   (23615),A      ;in LIST_SP
        LD   A,(23622)
        LD   (23616),A
        LD   A,(23610)      ;save ERR_NR error code
        LD   (23560),A      ;in LAST_K
        LD   BC,9000        ;new line number to jump to
        LD   (23662),BC     ;OLDPPC - CONTINUE line number
        LD   A,$ff
        LD   (23610),A      ;ERR_NR - reset error code
        INC  A
        LD   (23664),A      ;OSPCC  - CONTINUE statement number
        LD   HL,(23641)     ;E_LINE - address of editor buffer
        CALL $1652          ;ONE-SPACE - make a space in it
        EX   DE,HL
        LD   (HL),$e8       ;insert code for "CONTINUE"
MCLOOP: CALL $1f54          ;BREAK-KEY
        JP   C,$12cf        ;MAIN-3 - jump if not BREAK
        JR   MCLOOP         ;ignore BREAK

;++++++++++++++++++++++++++++++

;Change RAMTOP to "val" while preserving variables & GOSUBs
;POKE 23730,lo_val: POKE 23731,hi_val: RANDOMIZE USR ...
TOPRAM: LD HL, 18      ; 33  18  0       ;discard all but
        ADD HL, SP     ; 57              ;STMT_RET & (ERR_SP)
        LD SP, HL      ; 249             ;from m/c stack
        INC HL         ; 35
        LD B, H        ; 68
        LD C, L        ; 77
        INC HL         ; 35              ;loop to find end of
NEXT_3: INC HL         ; 35              ;old GOSUB stack
        INC HL         ; 35
        INC HL         ; 35
        LD A, (HL)     ; 126
        CP $3E         ; 254 62
        JR NZ, NEXT_3  ; 32  244
        PUSH HL        ; 229
        SBC HL, BC     ; 237 66
        LD B, H        ; 68
        LD C, L        ; 77
        DEC BC         ; 11
        DEC BC         ; 11
        POP HL         ; 225
        LD DE, (23730) ; 237 91  178 92  ;new RAMTOP
        LDDR           ; 237 184         ;copy old GOSUB stack
        INC DE         ; 19              ;to new location
        ex  de,hl      ; 235
        POP DE         ; 209
        LD DE,$1B76    ; 17  118 27      ;$1B76 STMT_RET
        POP BC         ; 193
        LD SP, HL      ; 249
        PUSH BC        ; 197
        LD (23613), SP ; 237 115 61  92  ;ERR_SP
        EX DE, HL      ; 235
        RES 4, (IY+1)  ; 253 203 1   166 ;+2A/+3/+3e fix
        JP (HL)        ; 233


;1216 LET hs=k0:
;     FOR i=k1 TO em: IF j$(i)<>"!" [OR v$(i)="!"] THEN GO TO 1224
;1218   LET st=FN s(y$(i))-k3*ir: LET le=CODE y$(i,k3)
;1219   FOR a=st TO st+(jp AND jp<=le)+(le AND jp>le)-k3 STEP k3:
;         LET loc=FN p(a+bo)
;1220     IF loc THEN IF j$(loc)<>"!" THEN
;           LET hs=hs+k1: LET sz=sz+k1: LET j$(loc)="!":
;           PRINT AT k3,k3*k8;sz;AT k4,k5;i,AT k4,kY-k6;loc,:
;           POKE a+bo+k2,PEEK (a+bo+k2)+k1:
;           IF loc=go OR go=k1 AND {u$(loc)>" "} THEN
;             LET b=CODE u$(loc)-kR: LET b=b-(b-cr AND b<k1):
;             PRINT AT k5,k0;"Found ";c$(b,k7 TO kY):
;             LET u$(loc)=CHR$ k0: LET so=loc: GO TO 1228
;1222   NEXT a: LET ht=ht+(hs OR jp<le): IF hs THEN GO TO 1216
;1223  [IF jp>le THEN LET v$(i)="!"]
;1224 NEXT i

;	 PRINT USR ms,sz,em,ir,jp,bo,go,cr,"y$(ir,6)","c$(cr,9)","s$(kH1,21),"u$(em,k2)","j$(em)",mcb;
;    returns: sz,so,ht,gt

ARRAYJ: equ  MCWORK     ;j$() location of address of row
ARRAYU: equ  MCWORK+2   ;u$() location of address of 1st row
ARRAYS: equ  MCWORK+4   ;s$()            "
ARRAYC: equ  MCWORK+6   ;c$()            "
ARRAYY: equ  MCWORK+8   ;y$()            "
;       equ  MCWORK+10,11

NUMBCR: equ  MCWORK+12  ;cr location (c$() rows)
NUMBGO: equ  MCWORK+14  ;go location (final key item destination)
NUMBBO: equ  MCWORK+16  ;bo location (y$() pointer offset)
NUMBJP: equ  MCWORK+18  ;jp location (search depth)
NUMBIR: equ  MCWORK+20  ;ir*3 location (y$() rows * 3)
NUMBEM: equ  MCWORK+22  ;em location (maximum location entries)
VALUSZ: equ  MCWORK+24  ;sz location (checked counter)

VALUHT: equ  MCWORK+26  ;ht location (hit count at this depth)
VALUHS: equ  MCWORK+27  ;hs location (hit count on this pass)
VALUSO: equ  MCWORK+28  ;so location (intermediate key item stop)
VALU_I: equ  MCWORK+30  ;i  location (outer loop counter)
VALUGT: equ  MCWORK+32  ;gt location (return GO TO line number)

NXTARR: push bc      ;1>
        call GET_A8     ;***get array address
        pop  bc      ;1<
        ld   a,c        ;HL=address of 1st row in a 2D array
        cp   b
        jr   nc,NOT1D
        dec  hl
        dec  hl         ;HL=address of 1st row in a 1D array
NOT1D:  ld   (ix+00),l
        ld   (ix+01),h
        inc  ix
        inc  ix
        djnz NXTARR
        ret

NXTINT: push bc      ;1>
        call FPTOBC     ;get integer parameter
        ld   (ix+00),c
        ld   (ix+01),b
        inc  ix
        inc  ix
        pop  bc      ;1<
        djnz NXTINT
        ret

MCBACK: ld   ix,MCWORK
        ld   bc,$0504
        call NXTARR     ;***
        inc  ix
        inc  ix
        ld   b,7
        call NXTINT     ;***
        ld   ix,MCWORK
        ld   de,(NUMBIR)
        ld   h,d
        ld   l,e
        add  hl,hl
        add  hl,de      ;HL=ir*3
        ld   (NUMBIR),hl
        ld   (ix+26),0  ;=ht

;1216 LET hs=k0:
GO1216: ld   (ix+27),0  ;=hs

;     FOR i=k1 TO em: IF j$(i)<>"!" [OR v$(i)="!"] THEN GO TO 1224
        ld   bc,1
FORIEM: ld   (VALU_I),bc
        ld   a,"!"
        ld   d,b
        ld   e,c
        ld   hl,(ARRAYJ)
        dec  de         ;DE=j$() row offset = (i-1)
        add  hl,de      ;HL=j$() row address
        cp   (hl)
G1224A: jp   nz,GO1224  ;***

;1218   LET st=FN s(y$(i))-k3*ir: LET le=CODE y$(i,k3)
        ld   h,d
        ld   l,e
        add  hl,de
        add  hl,de
        ex   de,hl      ;DE=y$() row offset = (i-1)*3
        ld   hl,(ARRAYY)
        add  hl,de      ;HL=y$() row address
        ld   e,(hl)
        inc  hl
        ld   d,(hl)     ;DE=FN s(y$(i))
        inc  hl
        ld   c,(hl)
        ld   b,0        ;BC=le=CODE y$(i,k3)
        ex   de,hl
        ld   de,(NUMBIR)
        sbc  hl,de      ;HL=st=FN s(y$(i))-k3*ir

;1219   FOR a=st TO st+(jp AND jp<=le)+(le AND jp>le)-k3 STEP k3:
        ld   a,(ix+18)  ;A =jp
        ld   (ix+19),c
        cp   c          ;C= le
        jr   z,JPLELE
        jr   nc,JPGTLE
JPLELE: ld   c,a        ;C =jp
JPGTLE: push hl      ;1>
        push bc      ;2>

;         LET loc=FN p(a+bo)
        ld   de,(NUMBBO)
        add  hl,de      ;HL=a+bo
        push hl      ;3>
        ld   e,(hl)
        inc  hl
        ld   d,(hl)     ;DE=loc=FN p(a+bo)

;1220     IF loc THEN IF j$(loc)<>"!" THEN
        ld   a,d
        or   e
G1222A: jp   z,GO1222   ;***
        ld   hl,(ARRAYJ)
        dec  de         ;DE=j$() row offset = (loc-1)
        add  hl,de      ;HL=j$() row address
        ld   a,"!"
        cp   (hl)
G1222B: jp   z,GO1222   ;***
        inc  de
        push de      ;4>;DE=loc

;           LET hs=hs+k1: LET sz=sz+k1: LET j$(loc)="!":
        inc  (ix+27)    ;hs=hs+k1
        ld   de,(VALUSZ)
        inc  de         ;sz=sz+k1
        ld   (VALUSZ),de
        ld   (hl),a     ;j$(loc)="!"

;           PRINT AT k3,k3*k8;sz;AT k4,k5;i,AT k4,kY-k6;loc,:
        ld   a,22
        rst  $10
        ld   a,3
        rst  $10
        ld   a,24
        rst  $10
        ld   bc,(VALUSZ)
        call OUTNUM     ;print sz
        ld   a,22
        rst  $10
        ld   a,4
        rst  $10
        ld   a,5
        rst  $10
        ld   bc,(VALU_I)
        call OUTNUM     ;print i
        ld   a,6
        rst  $10
        ld   a,22
        rst  $10
        ld   a,4
        rst  $10
        ld   a,19
        rst  $10
        pop  bc      ;4<;BC=loc
        push bc      ;4>
        call OUTNUM     ;print loc
        ld   a,6
        rst  $10

;           POKE a+bo+k2,PEEK (a+bo+k2)+k1
        pop  bc      ;4<;BC=loc
        pop  de      ;3<;DE=a+bo
        inc  de
        inc  de
        ex   de,hl
        inc  (hl)       ;POKE a+bo+k2,PEEK (a+bo+k2)+k1

;           IF loc=go OR go=k1 AND {u$(loc)<>"  " AND u$(loc)<>NULL} THEN
        ld   h,b
        ld   l,c        ;HL=loc
        push hl      ;3>
        ld   de,(NUMBGO);DE=go
        sbc  hl,de
        jr   z,FOUND1   ;IF loc=go
        dec  de
        ld   a,d
        or   e
        jr   nz,GO1222  ;go<>1
        ld   hl,(ARRAYU)
        dec  bc         ;BC=u$() row offset = (loc-1)
        add  hl,bc
        add  hl,bc      ;HL=u$() row address
        ld   a,(hl)
        inc  hl
        cp   (hl)
        jr   nz,FOUND1+1;u$(loc)<>"  " AND u$(loc)<>NULL
        or   $20
        cp   $20
        jr   nz,FOUND1+1;u$(loc)<>"  " AND u$(loc)<>NULL

;1222   NEXT a: LET ht=ht+(hs OR jp<le): IF hs THEN GO TO 1216
GO1222: pop  hl      ;3<;discard
        pop  bc      ;2<;BC=inner loop counter
        pop  hl      ;1<;HL=exit address
        inc  hl
        inc  hl
        inc  hl
        dec  c
        dec  c
        dec  c
NEXTA:  jp   nz,JPGTLE  ;***NEXT a

        xor  a
        cp   (ix+27)    ;cp hs
        push af      ;1>
        jr   nz,INC_HT
        ld   a,(ix+18)  ;A =jp
        cp   (ix+19)    ;cp le
        jr   nc,CHEKHS  ;jp>=le
INC_HT: inc  (ix+26)    ;ht=ht+(hs OR jp<le)
CHEKHS: pop  af      ;1<
        jp   nz,GO1216  ;***

;1224 NEXT i
GO1224: ld   bc,(VALU_I)
        inc  bc
        ld   hl,(NUMBEM)
        and  a
        sbc  hl,bc
NEXTI:  jp   nc,FORIEM  ;***NEXT i

        ld   bc,1225
        ld   (VALUGT),bc
        ret

FOUND1: dec  bc         ;BC=u$() row offset = (loc-1)
        ld   hl,(ARRAYU)
        add  hl,bc
        add  hl,bc      ;HL=u$() row address
        push hl      ;4>
;             LET b=CODE u$(loc)-kR: LET b=b-(b-cr AND b<k1):
        ld   a,(hl)
        inc  hl
        cp   (hl)
        jr   nz,FOUND2  ;u$(loc)<>"  " AND u$(loc)<>NULL
        or   $20
        cp   $20
        jr   z,FOUND3   ;u$(loc)="  " OR u$(loc)=NULL


;             PRINT AT k5,k0;"Found ";c$(b,k7 TO kY):
FOUND2: ld   b,(hl)
        dec  hl
        ld   c,(hl)     ;BC=c$() row#
        dec  bc
        call STAKBC
        ld   a,9
        call STAKA
        rst  $28
        defb $04,$38
        call FPTOBC     ;BC=c$() row offset = (b-1)*9
        ld   hl,(ARRAYC)
        add  hl,bc      ;HL=c$() row address
        ld   bc,6
        add  hl,bc      ;HL=c$(b,7) address

        ld   a,(hl)     ;A =s$() row number
        dec  a
        call STAKA
        ld   a,21
        call STAKA
        rst  $28
        defb $04,$38
        call FPTOBC     ;BC=s$() row offset = (a-1)*21
        ld   hl,(ARRAYS)
        add  hl,bc      ;HL=s$() row address

        ld   bc,19
FTEXT1: ld   de,FCDESC  ;***
        ldir            ;copy description to message area

        ld   a,22
        rst  $10
        ld   a,5
        rst  $10
        ld   a,0
        rst  $10
        ld   bc,25
FTEXT2: ld   de,FFOUND  ;***
        call PRINTS

;             LET u$(loc)=CHR$ k0: LET so=loc: GO TO 1228
FOUND3: pop  hl      ;4<;HL=u$(loc) address
        xor  a
        ld   (hl),a
        inc  hl
        ld   (hl),a     ;u$(loc)=NULL
        pop  de      ;3<;DE=loc
        pop  bc      ;2<;discard
        pop  hl      ;1<;discard
        ld   (VALUSO),de;so=loc
        ld   bc,1228
        ld   (VALUGT),bc
        ret

FFOUND: defm "Found "
FCDESC: defs 19


;++++++++++


;1232 LET st=FN s(i$(i)): LET le=CODE i$(i,k3): LET j$(i)="+":
;     IF i=sa THEN GO TO 1244
;1236 FOR a=st TO st+le-k3 STEP k3:
;       LET ex=FN p(a+eo): IF a<=FN s(a$(seq)) THEN GO TO 1242
;1238   IF j$(ex)="!" OR j$(ex)="-" THEN LET j$(ex)="+": GO TO 1244
;1240   IF j$(ex)="+" THEN LET j$(ex)="-"
;1242 NEXT a:
;     LET seq=seq-k1: LET a$(seq)=FN h$(k0): LET i=FN s(v$(seq)): GO TO 1232
;1244 LET v$(seq)=FN h$(i): LET a$(seq)=FN h$(a):
;     PRINT AT k8,k4;seq;AT k8,kA*k2;i,:
;     IF i<>sa THEN LET i=ex: LET seq=seq+k1: GO TO 1232

;	 PRINT USR ms,i,sa,eo,seq,"v$(em,2)","i$(ir,6)","a$(em,2)","j$(em)",mcf;
;    returns: nothing

ARRA_J: equ  MCWORK     ;j$() location of address of 1st row
ARRA_A: equ  MCWORK+2   ;a$()            "
ARRA_I: equ  MCWORK+4   ;i$()            "
ARRA_V: equ  MCWORK+6   ;v$()            "

NUMSEQ: equ  MCWORK+8   ;seq location (partial solution sequence)
NUM_EO: equ  MCWORK+10  ;eo location (i$() pointer offset)
NUM_SA: equ  MCWORK+12  ;sa location (search start)
NUM_I:  equ  MCWORK+14  ;i  location (from location)

VAL_EX: equ  MCWORK+16  ;ex location (exit location)

MCFRWD: ld   ix,MCWORK
        ld   bc,$0403
        call NXTARR     ;***
        ld   b,4
        call NXTINT     ;***
        ld   ix,MCWORK

;1232 LET st=FN s(i$(i)): LET le=CODE i$(i,k3): LET j$(i)="+":
GO1232: ld   de,(NUM_I)
        dec  de         ;DE=j$() row offset = (i-1)
        ld   hl,(ARRA_J)
        add  hl,de      ;HL=j$() row address
        ld   (hl),"+"   ;j$(i)="+"
        ld   h,d
        ld   l,e
        add  hl,de
        add  hl,de
        add  hl,hl
        ex   de,hl      ;DE=i$() row offset = (i-1)*6
        ld   hl,(ARRA_I)
        add  hl,de      ;HL=i$() row address
        ld   e,(hl)
        inc  hl
        ld   d,(hl)     ;DE=st=FN s(i$(i))
        inc  hl
        ld   c,(hl)
        ld   b,0        ;BC=le=CODE i$(i,k3)
        push bc      ;1>
        push de      ;2>

;     IF i=sa THEN GO TO 1244
        ld   de,(NUM_I)
        ld   hl,(NUM_SA)
        sbc  hl,de
        jr   z,GO1244
        pop  hl      ;2<;HL=exit address
        pop  bc      ;1<;BC=inner loop counter

;1236 FOR a=st TO st+le-k3 STEP k3:
GO1236: push bc      ;1>
        push hl      ;2>

;       LET ex=FN p(a+eo): IF a<=FN s(a$(seq)) THEN GO TO 1242
        ld   de,(NUM_EO)
        add  hl,de      ;HL=a+eo
        ld   e,(hl)
        inc  hl
        ld   d,(hl)     ;DE=ex=FN p(a+eo)
        ld   (VAL_EX),de
        ld   de,(NUMSEQ)
        dec  de
        ld   h,d
        ld   l,e
        add  hl,de
        ex   de,hl      ;DE=a$() row offset = (seq-1)*2
        ld   hl,(ARRA_A)
        add  hl,de      ;HL=a$() row address
        ld   e,(hl)
        inc  hl
        ld   d,(hl)     ;DE=FN s(a$(seq))
        pop  hl      ;2<;HL=a
        push hl      ;2>
        sbc  hl,de
        jr   c,GO1242   ;a<FN s(a$(seq))
        jr   z,GO1242   ;a=FN s(a$(seq))

;1238   IF j$(ex)="!" OR j$(ex)="-" THEN LET j$(ex)="+": GO TO 1244
        ld   de,(VAL_EX)
        dec  de         ;DE=j$() row offset = (ex-1)
        ld   hl,(ARRA_J)
        add  hl,de      ;HL=j$() row address
        ld   a,(hl)     ;A =j$(ex)
        cp   "!"
        jr   z,J_PLUS
        cp   "-"
        jr   nz,GO1240
J_PLUS: ld   (hl),"+"   ;j$(ex)="+"
        jr   GO1244

;1240   IF j$(ex)="+" THEN LET j$(ex)="-"
GO1240: cp   "+"
        jr   nz,GO1242
        ld   (hl),"-"   ;j$(ex)="-"

;1242 NEXT a:
;     LET seq=seq-k1: LET a$(seq)=FN h$(k0): LET i=FN s(v$(seq)): GO TO 1232
GO1242: pop  hl      ;2<;HL=exit address
        pop  bc      ;1<;BC=inner loop counter
        inc  hl
        inc  hl
        inc  hl
        dec  c
        dec  c
        dec  c
NEXTA2: jr   nz,GO1236  ;NEXT a

        ld   de,(NUMSEQ)
        dec  de
        ld   (NUMSEQ),de;seq=seq-k1
        dec  de
        ld   h,d
        ld   l,e
        add  hl,de
        ex   de,hl      ;DE=a$() row offset = (seq-1)*2
        ld   hl,(ARRA_A)
        add  hl,de      ;HL=a$() row address
        ld   (hl),0
        inc  hl
        ld   (hl),0     ;a$(seq)=FN h$(k0)
        ld   hl,(ARRA_V)
        add  hl,de      ;HL=v$() row address
        ld   e,(hl)
        inc  hl
        ld   d,(hl)
        ld   (NUM_I),de ;i=FN s(v$(seq))
        jp   GO1232     ;***

;1244 LET v$(seq)=FN h$(i): LET a$(seq)=FN h$(a):
GO1244: ld   de,(NUMSEQ)
        dec  de
        ld   h,d
        ld   l,e
        add  hl,de
        ex   de,hl      ;DE=v$() row offset = (seq-1)*2
        ld   hl,(ARRA_V)
        add  hl,de      ;HL=v$() row address
        ld   bc,(NUM_I)
        ld   (hl),c
        inc  hl
        ld   (hl),b     ;v$(seq)=FN h$(i)

        ld   hl,(ARRA_A)
        add  hl,de      ;HL=a$() row address
        pop  de      ;2<;DE=exit address
        pop  bc      ;1<;BC=inner loop counter
        ld   (hl),e
        inc  hl
        ld   (hl),d     ;a$(seq)=FN h$(a)

;     PRINT AT k8,k4;seq;AT k8,kA*k2;i,:
        ld   a,22
        rst  $10
        ld   a,8
        rst  $10
        ld   a,4
        rst  $10
        ld   bc,(NUMSEQ)
        call OUTNUM     ;print seq
        ld   a,22
        rst  $10
        ld   a,8
        rst  $10
        ld   a,20
        rst  $10
        ld   bc,(NUM_I)
        call OUTNUM     ;print i
        ld   a,6
        rst  $10

;     IF i<>sa THEN LET i=ex: LET seq=seq+k1: GO TO 1232
        ld   de,(NUM_I)
        ld   hl,(NUM_SA)
        and  a
        sbc  hl,de
        ret  z          ;i=sa

        ld   de,(VAL_EX)
        ld   (NUM_I),de ;i=ex
        ld   bc,(NUMSEQ)
        inc  bc
        ld   (NUMSEQ),bc;seq=seq+k1
NEXTI2: jp   GO1232     ;***


;++++++++++


;1251 FOR i=k1 TO em: LET loc=FN s(v$(i)): IF NOT loc THEN GO TO 1260
;1252   PRINT AT kA+k1,k0;seq,loc,: LET prq=FN s(a$(loc)): LET olq=seq:
;       IF prq THEN LET seq=prq
;1254   IF seq<olq THEN
;         FOR a=seq+k1 TO olq-k1: LET a$(FN s(w$(a)))=FN h$(k0): NEXT a
;1256   LET a$(loc)=FN h$(seq): LET w$(seq)=FN h$(loc)
;1258   LET seq=seq+k1:
;     NEXT i

;	 PRINT USR ms,em,seq,"v$(em,2)","a$(em,2)","w$(kK,2)",mcc;
;    returns: seq

ARRY_W: equ  MCWORK     ;w$() location of address of 1st row
ARRY_A: equ  MCWORK+2   ;a$()            "
ARRY_V: equ  MCWORK+4   ;v$()            "

NU_SEQ: equ  MCWORK+6   ;seq location (partial solution sequence)
NU_EM:  equ  MCWORK+8   ;em location (maximum location entries)

MCCOMP: ld   ix,MCWORK
        ld   bc,$0303
        call NXTARR     ;***
        ld   b,2
        call NXTINT     ;***
        ld   ix,MCWORK

;1251 FOR i=k1 TO em: LET loc=FN s(v$(i)): IF NOT loc THEN GO TO 1260
        ld   bc,1
FOR1EM: ld   (VALU_I),bc
        dec  bc
        ld   h,b
        ld   l,c
        add  hl,bc
        ex   de,hl      ;DE=v$() row offset = (i-1)*2
        ld   hl,(ARRY_V)
        add  hl,de      ;HL=v$() row address
        ld   c,(hl)
        inc  hl
        ld   b,(hl)     ;BC=loc=FN s(v$(i))
        ld   a,b
        or   c
        ret  z          ;IF NOT loc

;1252   PRINT AT kA+k1,k0;seq,loc,: LET prq=FN s(a$(loc)): LET olq=seq:
        push bc      ;1>
        ld   a,22
        rst  $10
        ld   a,11
        rst  $10
        ld   a,0
        rst  $10
        ld   bc,(NU_SEQ)
        call OUTNUM     ;print seq
        ld   a,6
        rst  $10
        pop  bc      ;1<;BC=loc
        push bc      ;1>
        call OUTNUM     ;print loc
        ld   a,6
        rst  $10

        pop  bc      ;1<;BC=loc
        push bc      ;1>
        dec  bc
        ld   h,b
        ld   l,c
        add  hl,bc
        ex   de,hl      ;DE=a$() row offset = (loc-1)*2
        ld   hl,(ARRY_A)
        add  hl,de      ;HL=a$() row address
        ld   c,(hl)
        inc  hl
        ld   b,(hl)     ;BC=prq=FN s(a$(loc))
        ld   de,(NU_SEQ);DE=olq=seq

;       IF prq THEN LET seq=prq
        ld   a,b
        or   c
        jr   z,GO1254   ;IF NOT prq
        ld   (NU_SEQ),bc;seq=prq

;1254   IF seq<olq THEN
GO1254: ld   hl,(NU_SEQ)
        and  a
        sbc  hl,de
        jr   nc,GO1256  ;IF seq>=olq

;         FOR a=seq+k1 TO olq-k1: LET a$(FN s(w$(a)))=FN h$(k0): NEXT a
        ld   bc,(NU_SEQ)
        dec  de         ;olq-k1
        push de      ;2>
NEXTA3: inc  bc         ;BC=a
        pop  de      ;2<
        ld   h,d
        ld   l,e
        and  a
        sbc  hl,bc
        jr   c,GO1256

        push de      ;2>
        dec  bc
        ld   h,b
        ld   l,c
        add  hl,bc
        ex   de,hl      ;DE=w$() row offset = (a-1)*2
        ld   hl,(ARRY_W)
        add  hl,de      ;HL=w$() row address
        ld   e,(hl)
        inc  hl
        ld   d,(hl)     ;DE=FN s(w$(a))
        dec  de
        ld   h,d
        ld   l,e
        add  hl,de
        ex   de,hl      ;DE=a$() row offset = (FN s(w$(a))-1)*2
        ld   hl,(ARRY_A)
        add  hl,de      ;HL=a$() row address
        ld   (hl),0
        inc  hl
        ld   (hl),0     ;a$(FN s(w$(a)))=FN h$(k0)
        inc  bc
        jr   NEXTA3     ;NEXT a

;1256   LET a$(loc)=FN h$(seq): LET w$(seq)=FN h$(loc)
GO1256: pop  bc      ;1<;BC=loc
        ld   d,b
        ld   e,c
        dec  de
        ld   h,d
        ld   l,e
        add  hl,de
        ex   de,hl      ;DE=a$() row offset = (loc-1)*2
        ld   hl,(ARRY_A)
        add  hl,de      ;HL=a$() row address
        ld   de,(NU_SEQ)
        ld   (hl),e
        inc  hl
        ld   (hl),d     ;a$(loc)=FN h$(seq)

        dec  de
        ld   h,d
        ld   l,e
        add  hl,de
        ex   de,hl      ;DE=w$() row offset = (seq-1)*2
        ld   hl,(ARRY_W)
        add  hl,de      ;HL=w$() row address
        ld   (hl),c
        inc  hl
        ld   (hl),b     ;w$(seq)=FN h$(loc)

;1258   LET seq=seq+k1:
        ld   bc,(NU_SEQ)
        inc  bc
        ld   (NU_SEQ),bc;seq=seq+k1

;     NEXT i
        ld   bc,(VALU_I)
        inc  bc
        ld   hl,(NU_EM)
        and  a
        sbc  hl,bc
NEXTI3: jp   nc,FOR1EM  ;***NEXT i
        ret


;++++++++++


;1260 FOR i=seq-k1 TO k1 STEP -k1:
;       LET fm=FN s(w$(i)): PRINT fm;" ";: LET h$(fm)=FN h$(bk)
;1261   LET ex=FN p(bk+k2): LET st=FN s(i$(fm)): LET le=CODE i$(fm,k3):
;       FOR a=st TO st+le-k3 STEP k3:
;         IF FN p(a+eo)<>ex THEN
;           NEXT a
;1262   IF a=st+le AND fm<em THEN
;         LET e$="EXIT ERROR "+STR$ fm+" > "+STR$ ex:
;         GO SUB e0: GO TO x0
;1263   POKE bk,FN g(fm): POKE bk+k1,FN h(fm): LET bk=bk-k2:
;     NEXT i: LET bk=bk+k2

;	 PRINT USR ms,eo,bk,em,seq,"i$(ir,6)","h$(em,2)","w$(kK,2)",mcs;
;    returns: bk, gt

ARRZ_W: equ  MCWORK     ;w$() location of address of 1st row
ARRZ_H: equ  MCWORK+2   ;h$()            "
ARRZ_I: equ  MCWORK+4   ;i$()            "

NU_BK:  equ  MCWORK+10  ;bk location (full solution location)
NU_EO:  equ  MCWORK+12  ;eo location (i$() pointer offset)

VAL_FM: equ  MCWORK+14  ;fm location (from location)

MCSAVE: ld   ix,MCWORK
        ld   bc,$0303
        call NXTARR     ;***
        ld   b,4
        call NXTINT     ;***
        ld   ix,MCWORK

;1260 FOR i=seq-k1 TO k1 STEP -k1:
        ld   bc,(NU_SEQ)
FOR2K1: dec  bc
        push bc      ;1>

;       LET fm=FN s(w$(i)): PRINT fm;" ";: LET h$(fm)=FN h$(bk)
        dec  bc
        ld   h,b
        ld   l,c
        add  hl,bc
        ex   de,hl      ;DE=w$() row offset = (i-1)*2
        ld   hl,(ARRZ_W)
        add  hl,de      ;HL=w$() row address
        ld   c,(hl)
        inc  hl
        ld   b,(hl)     ;BC=fm=FN s(w$(i))
        push bc      ;2>
        call OUTNUM     ;print fm
        ld   a," "
        rst  $10
        pop  bc      ;2<
        ld   (VAL_FM),bc

        dec  bc
        ld   h,b
        ld   l,c
        add  hl,bc
        ex   de,hl      ;DE=h$() row offset = (fm-1)*2
        push de      ;2>
        ld   hl,(ARRZ_H)
        add  hl,de      ;HL=h$() row address
        ld   de,(NU_BK)
        ld   (hl),e
        inc  hl
        ld   (hl),d     ;h$(fm)=FN h$(bk)

;1261   LET ex=FN p(bk+k2): LET st=FN s(i$(fm)): LET le=CODE i$(fm,k3):
        ex   de,hl      ;HL=bk
        inc  hl
        inc  hl
        ld   c,(hl)
        inc  hl
        ld   b,(hl)     ;BC=ex=FN p(bk+k2)
        ld   (VAL_EX),bc
        pop  de      ;2<;DE=h$() row offset = (fm-1)*2
        ld   h,d
        ld   l,e
        add  hl,de
        add  hl,de
        ex   de,hl      ;DE=i$() row offset = (fm-1)*6
        ld   hl,(ARRZ_I)
        add  hl,de      ;HL=i$() row address
        ld   e,(hl)
        inc  hl
        ld   d,(hl)     ;DE=st=FN s(i$(fm))
        inc  hl
        ld   c,(hl)
        ld   b,0        ;BC=le=CODE i$(fm,k3)
        ld   hl,(NU_EO)
        add  hl,de      ;HL=eo+st

;       FOR a=st TO st+le-k3 STEP k3:
FORAST: push bc      ;2>
        push hl      ;3>

;         IF FN p(a+eo)<>ex THEN
;           NEXT a
        ld   e,(hl)
        inc  hl
        ld   d,(hl)
        ex   de,hl      ;HL=FN p(a+eo)
        ld   bc,(VAL_EX)
        and  a
        sbc  hl,bc
        pop  hl      ;3<;HL=exit address
        pop  bc      ;2<;BC=inner loop counter
        jr   z,GO1263   ;IF FN p(a+eo)=ex

        inc  hl
        inc  hl
        inc  hl
        dec  c
        dec  c
        dec  c
        jr   nz,FORAST

;1262   IF a=st+le AND fm<em THEN
;         LET e$="EXIT ERROR "+STR$ fm+" > "+STR$ ex:
;         GO SUB e0: GO TO x0
GO1262: ld   hl,(VAL_FM)
        ld   bc,(NU_EM)
        and  a
        sbc  hl,bc
        jr   nc,GO1263

        pop  bc      ;1<;discard
        ld   bc,1262
        ld   (VALUGT),bc
        ret

;1263   POKE bk,FN g(fm): POKE bk+k1,FN h(fm): LET bk=bk-k2:
GO1263: ld   hl,(NU_BK)
        ld   bc,(VAL_FM)
        ld   (hl),c     ;POKE bk,FN g(fm)
        inc  hl
        ld   (hl),b     ;POKE bk+k1,FN h(fm)
        dec  hl
        dec  hl
        dec  hl
        ld   (NU_BK),hl ;bk=bk-k2

;     NEXT i: LET bk=bk+k2
        pop  bc      ;1<outer loop counter
        ld   a,c
        dec  a
        or   b
NEXTI4: jp   nz,FOR2K1  ;***

        ld   de,(NU_BK)
        inc  de
        inc  de
        ld   (NU_BK),de ;bk=bk+k2
        ld   bc,1264
        ld   (VALUGT),bc
        ret


;++++++++++


;PRINT USR ms,loc,"r$()","g$()",rrg; => let r$()=g$(loc)
;Transfer a variable length row from g$() to fixed length r$().
;Returns an empty record to r$() if an entry for loc is not found in g$().
READRG: call GET_A           ;***
        push hl           ;1>
        call GET_A           ;***
        pop  de           ;1<
        ld   bc,6
        add  hl,bc
        ld   (MCWORK+0),hl   ;r$() address
        ex   de,hl
        add  hl,bc
        ld   (MCWORK+2),hl   ;g$() address
        push hl           ;1>
        call FPTOBC          ;BC=location#
        pop  hl           ;1<;HL=g$() address

SWAP_1: ld   (MCWORK+5),hl   ;=next g$() record address
        ld   e,(hl)
        ld   a,e
        inc  hl
        ld   d,(hl)          ;DE=g$() location#
        add  a,d
        jr   nz,SWAP_2

CLEARR: ld   hl,(MCWORK+0)   ;no g$() row for this location#, so ...
        ld   (hl),a          ;clear r$() & return
        ld   d,h
        ld   e,l
        inc  de
        ld   bc,127
        ldir
        ret

SWAP_2: ex   de,hl
        and  a
        sbc  hl,bc           ;(current location)-(requested location)
        ex   de,hl
        push af

        inc  hl
        ld   a,(hl)          ;=required count (total)
        inc  hl
        inc  hl
        add  a,(hl)          ;+set      count (total)
        inc  hl
        inc  hl
        add  a,(hl)          ;+reset    count (total)
        sla  a               ;(entry counts)*2 = (g$() detail length)
        inc  hl
        inc  hl
        ld   d,0
        ld   e,a             ;DE=g$() detail length
        pop  af              ;F =flags for "sbc  hl,bc"
        jr   z,EXPG2R
        add  hl,de           ;current g$() loc# doesn't match
        jr   SWAP_1          ;back to check next g$() record

        ;expand g$(loc) record to fit r$()
EXPG2R: call CLEARR          ;*** clear r$()
        exx
        ld   hl,(MCWORK+5)   ;HL=start of g$(loc) record
        ld   de,(MCWORK+0)   ;DE=start of r$() row
        push hl           ;1>
        ld   bc,8
        ldir                 ;copy header g$()->r$()
        exx
        pop  hl           ;1<;HL=start of source row
        inc  hl
        inc  hl
        ld   b,3
        ld   d,0
EXPBAK: ld   e,(hl)          ;E =entry count  (0-20)
        sla  e               ;DE=entry length (0-40)
        push de           ;1>
        exx
        pop  bc           ;1<
        ld   a,40
        sub  c
        cp   40
        jr   z,EXPNON        ;empty entry; skip to next
        ldir
EXPNON: ld   c,a             ;BC=40-(length copied)
        ex   de,hl           ;if destination r$()
        add  hl,bc           ;skip over unused part of r$() entry
        ex   de,hl
        exx
        inc  hl
        inc  hl
        djnz EXPBAK
        ret

;       org  MCWORK
;       defw 0000            ;(ix+00) r$() row address
;       defw 0000            ;(ix+02) g$() row address
;       defb 00              ;(ix+04) destination array
;       defw 0000            ;(ix+05) g$() record address


;++++++++++


LABELS0:
        defw ADJUST
        defw STACK
        defw PUT_T
        defw GET_T
;       defw MCMOVE
;       defw MCUPDT
        defw MCINIT
        defw COPYAB
        defw MCTRAP

LABELS1:
       ;defw FINDAR
        defw TOPRAM
       ;defw INITRP
        defw MCBACK
        defw MCFRWD
        defw MCCOMP
        defw MCSAVE
        defw READRG