
;Shared Machine Code Routines
;============================

;Note that ORG addresses are for illustrative purposes, as actual locations
;vary depending upon the length of each program, each machine code block
;being positioned between the end of BASIC and the start of VARS.


LOCATE: equ  25000           ;CLEAR LOCATE-1
STACKP: equ  23670           ;=SEED =where SP is saved
        org  LOCATE

;adjust any absolute addresses
ADJUST: push bc
        pop  hl              ;HL=actual runtime location of ADJUST
        ld   de,LOCATE       ;DE=assembled      location of ADJUST
        and  a
        sbc  hl,de
        push hl
        pop  bc              ;BC=runtime displacement from LOCATE
        ret  z               ;return if no displacement

PUT_SP: di
        ld   (STACKP),sp     ;save stack pointer
        ld   hl,CHANGE
        add  hl,bc           ;HL=CHANGE+displacement
        ld   sp,hl           ;point SP to list of changes
ADLOOP: pop  hl
        ld   a,h
        or   l
        jr   z,GET_SP

        add  hl,bc           ;HL=location+displacement
        ex   de,hl
        pop  hl
        add  hl,bc           ;HL=content+displacement
        ex   de,hl           ;HL=address of location to change
        ld   (hl),e          ;DE=new content for that location
        inc  hl
        ld   (hl),d
        jr   ADLOOP

GET_SP: ld   sp,(STACKP)     ;restore stack pointer
        ei
        ret

        ;address relocation table
        ;terminated by 0000
        ;part 0: shared routines
        ;ld   hl,$nnnn assembles as 2A nn nn rather than ED 6B nn nn
CHANGE: defw PUT_T+1,SET_W,   GET_T+1,SET_W,    SET_W+1,GET_A,    GET_A+4,FIND
        defw GET_A8+1,GET_A,  MCINIT+1,GET_A8
        defw LABELS0,ADJUST,  LABELS0+2,STACK,  LABELS0+4,PUT_T,  LABELS0+6,GET_T
        defw LABELS0+8,MCINIT,LABELS0+10,MCTRAP

        ;part 4: AGES-4 routines
        defw WINDOZ+1,GETVJV, WINDOZB+1,POS2LOC,ADJUSTP+1,ADJUSTV,GETVJV+1,GET_A,  GETVJJ+1,GET_A8
        defw GETPOS+1,GETVJV, GETPOS1+1,GET_A
        defw GETPOS2+1,GETLEX,GETPOS2+4,POS2LOC,GETPOS2+7,POS2EX
        defw TXTPIC+1,GETPOS, ADJUSTT+1,ADJUSTV,EDITEXT+1,GETPOS
        defw LABELS1,WINDOZ,  LABELS1+2,TXTPIC, LABELS1+4,EDITEXT

        ;end marker
        defw $0000

CLOWER: equ  $0d6e           ;CLS-LOWER clear lower screen
CHOPEN: equ  $1601           ;CHAN-OPEN open channel A
OUTNUM: equ  $1a1b           ;OUT-NUM-1 print integer BC 0-9999
PRINTS: equ  $203c           ;PR-STRING print a string length BC from DE
PLOT:   equ  $22dc           ;PLOT      plot a point
DRAW:   equ  $2477           ;LINE-DRAW draw a line
STORES: equ  $2ab6           ;to calculator stack: AEDCB
FETCHS: equ  $2bf1           ;from calculator stack: AEDCB
STAKA:  equ  $2d28           ;to calculator stack: A
STAKBC: equ  $2d2b           ;to calculator stack: BC
FPTOBC: equ  $2da2           ;from calculator stack: BC
FPTOA:  equ  $2dd5           ;from calculator stack: A

K_CUR:  equ  23643
CH_ADD: equ  23645
WORKSP: equ  23649
STKBOT: equ  23651
MCWORK: equ  65368

;stack function parameters passed from BASIC by:
;PRINT USR mcstack,arg1[,arg2...,argn],addr;
STACK:  ld   hl,14
        add  hl,sp
        ld   sp,hl           ;undo USR
        ld   ix,0
GETARG: rst  $20
        call $24fb           ;SCANNING
        inc  ix
        rst  $18
        cp   ","
        jr   z,GETARG
        call FPTOBC          ;BC=address
        ld   h,b
        ld   l,c
        push ix
        pop  bc
        dec  bc
        ld   a,c             ;A =number of parameters
        ld   ix,MCWORK
        jp   (hl)            ;JP addr


;find a string array; call with DE=address of array name
FIND:   ld   bc,(CH_ADD)
        push bc
        ld   (CH_ADD),de
        call $28b2           ;LOOK-VARS
        pop  bc
        ld   (CH_ADD),bc
        ret

;	PRINT USR mcstack,"t$()",t;
;     t$()=array name,
;     t=address of machine code routine
;      =mcputt : put a copy of header array named in t$ into workspace
;      =mcgett : get a copy of header array named in t$ from workspace

;put a copy of header array t$ into workspace
PUT_T:  call SET_W           ;***
        ldir
        ret
;get a copy of header array t$ from workspace
GET_T:  call SET_W           ;***
        ex   de,hl
        ldir
        ret

;set header array & workspace addresses
SET_W:  call GET_A           ;***HL=array address
        ld   de,65536-168    ;DE=workspace address
        ld   bc,54           ;BC=header array size (=48 + 6 byte header)
        ret

;get array address
GET_A:  call FETCHS          ;DE=array name address
        call FIND            ;***HL=array address
        ret
GET_A8: call GET_A           ;***HL=array address
        ld   bc,8
        add  hl,bc           ;HL=address of 1st row in a 2D array
        ret

;get a +ve/-ve integer off the calculator stack
CHKNEG: call FPTOBC
        ret  z               ;return if not -ve
        ld   hl,0000
        and  a
        sbc  hl,bc           ;negate BC
        ld   b,h
        ld   c,l
        ret


;PRINT USR mcstack,l,x,n$,mcinitn;
MCINIT: call GET_A8         ;***
        push hl             ;HL=address of 1st row in array
        call FPTOBC         ;BC=number of rows
        push bc
        call FPTOA          ;A =row length
        ld   b,a
        ld   c,a
        pop  de             ;DE=number of rows
        pop  hl
        push hl             ;save original array pointer
REPS:   ld   (hl),$00       ;initialise 1st row
        inc  hl
        djnz REPS
        push hl             ;save updated array pointer
        dec  de             ;DE=(number of rows)-1 for row already done
        ld   hl,0000
        ld   b,c
REPD:   add  hl,de          ;multiply (row#-1)*rowlen
        djnz REPD
        ld   b,h
        ld   c,l            ;BC=number of bytes to copy
        pop  de             ;DE=starting destination (1+rowlen byte of array)
        pop  hl             ;HL=starting source      (1st byte of array)
        ldir                ;initialise remainder of array rows
        ret


;ERROR TRAP - machine code segment

MCTRAP: RES  4,(IY+1)       ;patch for +3 128k mode

        LD   A,(23621)      ;save PPC line number (where error occurred)
        LD   (23615),A      ;in LIST_SP
        LD   A,(23622)
        LD   (23616),A
        LD   A,(23610)      ;save ERR_NR error code
        LD   (23560),A      ;in LAST_K
        LD   BC,9000        ;new line number to jump to
        LD   (23662),BC     ;OLDPPC - CONTINUE line number
        LD   A,$ff
        LD   (23610),A      ;ERR_NR - reset error code
        INC  A
        LD   (23664),A      ;OSPCC  - CONTINUE statement number
        LD   HL,(23641)     ;E_LINE - address of editor buffer
        CALL $1652          ;ONE-SPACE - make a space in it
        EX   DE,HL
        LD   (HL),$e8       ;insert code for "CONTINUE"
MCLOOP: CALL $1f54          ;BREAK-KEY
        JP   C,$12cf        ;MAIN-3 - jump if not BREAK
        JR   MCLOOP         ;ignore BREAK

;++++++++++++++++++++++++++++++

PUTGET: equ  MCWORK          ;PUT=+1, GET=0, DEL=-1
ROW1:   equ  MCWORK+1
COL1:   equ  MCWORK+2
ROW2:   equ  MCWORK+3
COL2:   equ  MCWORK+4
T_PSIZE:equ  MCWORK+5        ;new text/pict item byte size
POINTER:equ  MCWORK+7        ;v$() pointer for changed entry
J_DATA: equ  MCWORK+9        ;j$() data start address
V_DATA: equ  MCWORK+11       ;v$() data start address
V_ROW:  equ  MCWORK+13       ;v$() address for changed entry
T_PADDR:equ  MCWORK+15
WIDTH:  equ  MCWORK+17
T_PTYPE:equ  MCWORK+19       ;"t"=text; "p"=pict
BLOCK:  equ  MCWORK+29       ;current data block# (POKEd from BASIC)

DF_CC:  equ  23684
V_SIZE: equ  7500

;PRINT USR ms,mact,r1,c1,r2,c2,sz,js,"j$()","v$()",mwin;;
;PUT/GET a screen segment/window
WINDOZ: call GETVJV          ;***
        call FPTOBC
        ld   (T_PSIZE),bc
        ld   hl,COL2
        ld   b,5
WINPAR: push bc
        push hl
        call FPTOA
        jr   z,NOTMIN
        neg
NOTMIN: pop  hl
        ld   (hl),a
        dec  hl
        pop  bc
        djnz WINPAR

        ld   a,(ROW1)
        ld   c,a
        ld   a,(ROW2)
        sub  c
        inc  a
        ld   (ROW2),a        ;(ROW2)=window depth
        ld   a,(COL1)
        ld   c,a
        ld   a,(COL2)
        sub  c
        inc  a
        ld   (WIDTH),a
        xor  a
        ld   (WIDTH+1),a     ;(WIDTH)=window width

WINDOZB:call POS2LOC         ;***
        ld   b,a             ;B =exits
        add  hl,de           ;skip over locntext
SKIPEX: inc  hl
        inc  hl              ;skip over exit#
        ld   e,(hl)
        inc  hl
        ld   d,(hl)          ;DE=exittextbytes
        inc  hl
        add  hl,de           ;skip over exittext
        djnz SKIPEX
        ld   c,(hl)
        inc  hl
        ld   b,(hl)          ;BC=old pictbytes
        inc  hl
        inc  hl              ;skip over pict coordinates
        inc  hl
        inc  hl
        inc  hl
        ld   (T_PADDR),hl

        ld   a,(PUTGET)
        or   a
        jr   nz,WINPUT

WINGET: ld   hl,(T_PADDR)  ;*ld   de,(T_PADDR)
GETWIN: push hl            ;*push de
        ld   a,22            ;AT
        rst  $10
        ld   a,(ROW1)
        rst  $10
        ld   a,(COL1)
        rst  $10
        ld   de,(DF_CC)    ;*ld   hl,(DF_CC);DE=display file address of PRINT position

        pop  hl            ;*pop  de
        push de            ;*push hl
        ld   a,d           ;*ld   a,h
        rrca
        rrca
        rrca
        and  $03
        or   $58
        ld   d,a           ;*ld   h,a
        ld   bc,(WIDTH)
        ldir                 ;copy attributes for this row
        pop  de            ;*pop  hl

        ld   a,8
GETNEXT:                     ;copy current row of width*8 bytes
        ld   bc,(WIDTH)
        push de            ;*push hl
        ldir
        ex   de,hl         ;*nop
        pop  hl
        ld   bc,256
        add  hl,bc
        ex   de,hl         ;*nop
        dec  a
        jr   nz,GETNEXT
        push hl            ;*nop
        ld   hl,ROW1
        inc  (hl)
        inc  hl
        inc  hl
        dec  (hl)            ;ROW2 counter
        pop  hl            ;*nop
        jr   nz,GETWIN
        ret

WINPUT: add  hl,bc           ;HL=address of next v$() segment
        ex   de,hl
        ld   hl,(T_PSIZE)    ;HL=new pictbytes
        sbc  hl,bc           ;HL=new pictbytes - old pictbytes
        ld   (ix+19),"p"
ADJUSTP:call nz,ADJUSTV      ;***make/reclaim room for picture
PUTWIN0:ld   de,(T_PADDR)
        ld   bc,(T_PSIZE)
        ld   a,b
        or   c
        ret  z               ;picture was deleted

PUTWIN1:push de
        ld   a,22            ;AT
        rst  $10
        ld   a,(ROW1)
        rst  $10
        ld   a,(COL1)
        rst  $10
        ld   hl,(DF_CC)      ;DE=display file address of PRINT position
        pop  de

        push hl
        ld   a,h
        rrca
        rrca
        rrca
        and  $03
        or   $58
        ld   h,a
        ld   bc,(WIDTH)
        ldir                 ;copy attributes for this row
        pop  hl

        ld   a,8
PUTNEXT:                     ;copy current row of width*8 bytes
        ld   bc,(WIDTH)
        push hl
        ldir
        pop  hl
        ld   bc,256
        add  hl,bc
        dec  a
        jr   nz,PUTNEXT
        ld   hl,ROW1
        inc  (hl)
        inc  hl
        inc  hl
        dec  (hl)            ;ROW2 counter
        jr   nz,PUTWIN1
        ret

        ;make/reclaim room for text/picture
ADJUSTV:push de           ;1>;DE=address of next v$() segment
        push hl           ;2>;HL=new-old databytes difference
        ld   hl,(V_ROW)      ;v$() address for changed entry
FINDEND:ld   c,(hl)          ;find end of v$() data
        inc  hl
        ld   b,(hl)          ;BC=length-2
        inc  hl
        add  hl,bc
        ld   a,b
        or   c
        jr   nz,FINDEND

        inc  hl              ;skip end of data marker
        inc  hl
        ld   d,h
        ld   e,l             ;DE=v$() data end address (1st free byte)
        ld   bc,(V_DATA)     ;BC=v$() data start address
        sbc  hl,bc           ;HL=v$() end of data pointer (last used byte)
        pop  bc           ;2<;BC=new-old databytes difference
        push bc           ;2>
        add  hl,bc
        ld   bc,V_SIZE       ;BC=DIM v$()
        sbc  hl,bc
        jp   nc,$2a20        ;"3 - Subscript wrong"

        ;make/reclaim room for text/picture
        pop  bc           ;2<
        pop  hl           ;1<
        push bc           ;1>
        ex   de,hl           ;DE=address of next v$() segment
        and  a
        sbc  hl,de           ;HL=(data end address)-(next segment address)
        ld   b,h
        ld   c,l             ;BC=bytes to move
        pop  hl           ;1<;HL=new-old databytes difference
        push hl           ;1>
        ld   a,b
        or   c
        jr   z,PUTDIFF       ;nothing to move
        add  hl,de
        ex   de,hl           ;next record address: HL=old; DE=new
        jr   nc,BIGGER
        ldir                 ;reclaim room
        jr   PUTDIFF
BIGGER: add  hl,bc
        ex   de,hl
        add  hl,bc
        ex   de,hl
        dec  hl
        dec  de
        lddr                 ;make room

PUTDIFF:pop  bc           ;1<;BC=new-old databytes difference
        ld   hl,(T_PADDR)
        dec  hl

        ld   a,"t"
        cp   (ix+19)
        jr   z,NOCOORD
        dec  hl              ;skip over pict coordinates
        dec  hl
        dec  hl
        dec  hl

NOCOORD:ld   d,(hl)
        dec  hl
        ld   e,(hl)          ;DE=old databytes
        ex   de,hl
        add  hl,bc           ;HL=new databytes
        ex   de,hl
        ld   (hl),e
        inc  hl
        ld   (hl),d
        inc  hl

        cp   (ix+19)
        jr   z,COORDNO
        push bc           ;1>
        ex   de,hl           ;DE=pict coord address in v$() record
        ld   hl,ROW1
        ld   bc,4
        ldir                 ;update pict coordinates
        pop  bc           ;<1;BC=byte length difference

COORDNO:ld   hl,(V_ROW)      ;HL=;v$() address for changed entry
        ld   e,(hl)
        inc  hl
        ld   d,(hl)          ;DE=old record length
        ex   de,hl
        add  hl,bc           ;HL=new record length
        ex   de,hl
        ld   (hl),d
        dec  hl
        ld   (hl),e

        ld   de,(POINTER)    ;DE=v$() pointer for changed entry
        push bc           ;1>
        exx
        pop  de           ;1<;D'E'=byte length difference
        ld   hl,(J_DATA)
        push hl           ;1>
        dec  hl
        dec  hl
        dec  hl
        ld   b,(hl)
        dec  hl
        ld   c,(hl)          ;B'C'=number of rows in j$()
        exx
        pop  hl           ;1<;HL=j$() current address

        ;loop to update j$() pointers to v$() records above changed record
CHECKJ: ld   c,(hl)
        inc  hl
        ld   b,(hl)          ;BC=j$() pointer to v$()
        inc  hl
        ld   a,(BLOCK)
        cp   (hl)
        jr   nz,NEXT_J2      ;location not in current block
        push de           ;1>
        ex   de,hl           ;HL=v$() pointer for changed entry
        and  a               ;DE=j$() current address
        sbc  hl,bc           ;HL=(v$() pointer for changed entry)-(j$() pointer to v$())
        jr   nc,NEXT_J1
        push bc           ;2>
        exx
        pop  hl           ;2<;H'L'=j$() pointer to v$()
        add  hl,de
        push hl           ;2>
        exx
        ex   de,hl           ;HL=j$() current address
        pop  de           ;2<;DE=adjusted j$() pointer to v$()
        dec  hl
        ld   (hl),d
        dec  hl
        ld   (hl),e
        inc  hl
        inc  hl
        ex   de,hl

NEXT_J1:ex   de,hl           ;HL=j$() current address
        pop  de           ;1<;DE=v$() pointer for changed entry
NEXT_J2:inc  hl
        inc  hl
        exx
        dec  bc
        ld   a,b
        or   c
        exx
        jr   nz,CHECKJ
        ret

GETVJV: call GET_A           ;***
        ld   bc,6
        add  hl,bc
        ld   (V_DATA),hl
GETVJJ: call GET_A8          ;***
        ld   (J_DATA),hl
        call FPTOBC
        ld   (POINTER),bc
        ret

GETLEX: call FPTOBC
        ld   (EXITN),bc
        call FPTOBC
        ld   (LOCATN),bc
        call FPTOA
        jr   z,NOTNEG
        neg
NOTNEG: ld   (ACTION),a      ;A =action: +1=update,-1=delete
        ret

POS2LOC:ld   hl,(V_DATA)
        ld   de,(POINTER)
        add  hl,de           ;HL=2nd byte address in entry
        dec  hl
        ld   (V_ROW),hl
        inc  hl
        inc  hl
        ld   a,(hl)          ;A =exits
        inc  hl
        inc  hl
        inc  hl
        ld   e,(hl)
        inc  hl
        ld   d,(hl)          ;DE=locntextbytes
        inc  hl
        ret

POS2EX: ex   af,af'
        ld   bc,(EXITN)
        ld   a,b
        or   c
        ret  z               ;stay at locntext if no exit#
        ex   af,af'
        or   a
        jp   z,$288b         ;"Q - Parameter error"
        add  hl,de           ;skip over locntext
CHECKEX:ld   e,(hl)
        inc  hl
        ld   d,(hl)          ;DE=exit#
        inc  hl
        ex   de,hl
        and  a
        sbc  hl,bc           ;HL=exit# - (EXITN)
        ex   de,hl
        ld   e,(hl)
        inc  hl
        ld   d,(hl)          ;DE=exittextbytes
        inc  hl
        ret  z
        add  hl,de           ;skip over exittext
        dec  a
        jr   nz,CHECKEX
        jp   $288b           ;"Q - Parameter error"

GETPOS: call GETVJV          ;***
        call FETCHS
        ld   (L_NAME),de
        call STORES
GETPOS1:call GET_A           ;***
        ld   (L_HEAD),hl
GETPOS2:call GETLEX          ;***
        call POS2LOC         ;***
        call POS2EX          ;***
        ret

;++++++++++++++++++++++++++++++

;    REM v$: length-2 + exits
;    rem   + locn# + locntextbytes + locntext
;    rem   + exit# + exittextbytes + exittext [+ exit# + exittextbytes + exittext ...]
;    rem   + pictbytes + r1+c1+r2+c2 [+ picture]

;PRINT USR ms,+k1,loc,k0,  "l$:",js,"j$()","v$()",mtp;: REM update location text in v$()
;PRINT USR ms,+k1,loc,exit,"l$:",js,"j$()","v$()",mtp;: REM update exit text in v$()
;PRINT USR ms,-k1,loc,k0,  "l$:",js,"j$()","v$()",mtp;: REM delete location text from v$()
;PRINT USR ms,-k1,loc,exit,"l$:",js,"j$()","v$()",mtp;: REM delete exit text from v$()

ACTION: equ  MCWORK+20
LOCATN: equ  MCWORK+21       ;location#
EXITN:  equ  MCWORK+23       ;exit#
L_HEAD: equ  MCWORK+25       ;l$ header address
L_NAME: equ  MCWORK+27       ;l$ "l$:"  address

TXTPIC: call GETPOS          ;***HL points to original text in v$(); DE=length
        ld   (T_PADDR),hl
        ld   b,d
        ld   c,e
        add  hl,bc           ;HL=address of next v$() segment in record
        ld   a,(ACTION)
        cp   $ff
        jr   nz,TXTUPD
        ld   de,0000         ;delete text; set new size=0
        jr   TXTDIF
TXTUPD: push hl
        ld   hl,(L_HEAD)
        inc  hl
        ld   e,(hl)
        inc  hl
        ld   d,(hl)          ;DE=l$ length
        pop  hl
TXTDIF: and  a
        ex   de,hl           ;DE=address of next v$() segment in record
        ld   (T_PSIZE),hl
        sbc  hl,bc           ;HL=new textbytes - old textbytes
        ld   (ix+19),"t"
ADJUSTT:call nz,ADJUSTV      ;***make/reclaim room for text
        ld   de,(T_PADDR)
        ld   bc,(T_PSIZE)
        ld   a,b
        or   c
        ret  z               ;text deleted
        ld   hl,(L_HEAD)
        inc  hl
        inc  hl
        inc  hl
        ldir
        ret

;++++++++++++++++++++++++++++++

;PRINT USR ms,+k1,loc,k0,  "l$:",js,"j$()","v$()",mut;: REM edit location text in v$()
;PRINT USR ms,+k1,loc,exit,"l$:",js,"j$()","v$()",mut;: REM edit exit text in v$()

DF_SZ:  equ  23659

EDITEXT:call GETPOS          ;***HL points to original text in v$(); DE=length
        ld   a,$01           ;open channel 'K'
        call $1601           ;CHAN_OPEN
        call $0d6e           ;CLS_LOWER
        ld   hl,(L_HEAD)
        inc  hl
        ld   c,(hl)
        inc  hl
        ld   b,(hl)          ;BC=l$ length
        inc  hl
        push hl
        push bc
        call $16bf           ;SET_WORK (reset workspace)
        pop  bc
        push bc
        inc  bc
        call $0030           ;BC_SPACES (make room for l$)
        pop  bc
        pop  hl
        ld   de,(WORKSP)
        ld   a,b
        or   c
        jr   z,NOTEXT
        ldir                 ;copy l$ to INPUT workspace
NOTEXT: ex   de,hl
        ld   (hl),$0d        ;followed by ENTER
        ld   (K_CUR),hl
        ld   bc,(CH_ADD)
        push bc
        ld   de,(L_NAME)
        ld   (CH_ADD),de
        call $1c1f           ;CLASS_01
        set  7,(iy+$37)      ;FLAGX: "using INPUT LINE"
        res  6,(iy+$37)      ;FLAGX: "string result"
        set  5,(iy+$37)      ;FLAGX: "INPUT mode"
        call $0f2c           ;EDITOR
        res  5,(iy+$37)      ;FLAGX: "edit mode"

        ld   hl,(STKBOT)
        ld   de,(WORKSP)
        scf
        sbc  hl,de
        ld   b,h
        ld   c,l
        call $2ab2           ;STK_ST_$
        ld   de,(L_NAME)
        ld   (CH_ADD),de
        call $2aff           ;LET (l$=edited text)
        pop  bc
        ld   (CH_ADD),bc
        ld   a,2
        ld   (DF_SZ),a       ;reset lower screen to 2 lines
        call $0d6e           ;CLS_LOWER
        res  0,(iy+02)       ;TV_FLAG
        ret


;++++++++++++++++++++++++++++++

LABELS0:defw ADJUST, STACK,  PUT_T,  GET_T,  MCINIT, MCTRAP
LABELS1:defw WINDOZ, TXTPIC, EDITEXT
MCSIZE: defw MCSIZE-ADJUST+3
        defb 10              ;number of entries in this segment